"""
    domonic.templates
    ====================================
    some builtin templates

"""

from __future__ import annotations

import os
import platform
import sys
from collections.abc import Iterable

from domonic.constants import http_response_status_codes
from domonic.html import *


def _normalize_children(content) -> list:
    if content is None:
        return []
    if isinstance(content, (str, bytes)):
        return [content]
    if isinstance(content, tuple):
        return list(content)
    if isinstance(content, Iterable):
        return list(content)
    return [content]


class _page_template:
    def __init__(self, title_text: str, body_children=None, wholepage: bool = True):
        self.title_text = title_text
        self.wholepage = wholepage
        self.body_node = body(*_normalize_children(body_children))
        self.page = html(head(title(self.title_text), meta(_charset="utf-8")), self.body_node)
        self.content = self.page if wholepage else self.body_node

    def __str__(self) -> str:
        return f"{self.content}"


class status_page(_page_template):
    DEBUG_MODE = False

    def __init__(self, code: int = 404, wholepage: bool = True):
        self.status_code = code
        self.message = http_response_status_codes[code]

        self.status_node = div(_id="status")
        self.status_node += h1(f"{self.status_code}")
        self.status_node += p(self.message)

        super().__init__(f"{self.status_code}", self.status_node, wholepage)


class blank_page(_page_template):
    def __init__(self, title_text: str = "Untitled", content=None, wholepage: bool = True):
        super().__init__(title_text, content, wholepage)


class message_page(_page_template):
    def __init__(
        self,
        title_text: str = "Message",
        heading: str = "Message",
        message: str = "",
        detail: str | None = None,
        wholepage: bool = True,
    ):
        content = div(_id="message")
        content += h1(heading)
        if message:
            content += p(message)
        if detail:
            content += pre(detail)
        super().__init__(title_text, content, wholepage)


class redirect_page(_page_template):
    def __init__(
        self,
        url: str,
        delay: int = 0,
        message: str | None = None,
        wholepage: bool = True,
    ):
        self.url = url
        self.delay = delay
        self.message = message or f"Redirecting to {url}"
        self.wholepage = wholepage

        self.body_node = body(
            div(
                h1("Redirecting"),
                p(self.message),
                p(a(url, _href=url)),
                _id="redirect",
            )
        )
        self.page = html(
            head(
                title("Redirecting"),
                meta(_charset="utf-8"),
                meta(_http_equiv="refresh", _content=f"{delay};url={url}"),
            ),
            self.body_node,
        )
        self.content = self.page if wholepage else self.body_node


class maintenance_page(_page_template):
    def __init__(
        self,
        retry_after: str | None = None,
        message: str = "The service is temporarily unavailable while we carry out maintenance.",
        wholepage: bool = True,
    ):
        content = div(_id="maintenance")
        content += h1("Service Unavailable")
        content += p(message)
        if retry_after:
            content += p(f"Retry after: {retry_after}")
        super().__init__("503", content, wholepage)


class runtime_page(_page_template):
    def __init__(
        self,
        title_text: str = "Runtime Information",
        include_environment: bool = False,
        include_cmd: bool = False,
        wholepage: bool = True,
    ):
        rows = [
            tr(th("Python"), td(sys.version.split()[0])),
            tr(th("Executable"), td(sys.executable)),
            tr(th("Platform"), td(platform.platform())),
            tr(th("Implementation"), td(platform.python_implementation())),
            tr(th("Working Directory"), td(os.getcwd())),
            tr(th("Process ID"), td(str(os.getpid()))),
        ]

        if include_cmd:
            try:
                from domonic.cmd import hostname, whoami

                rows.append(tr(th("Host Name"), td(str(hostname()).strip())))
                rows.append(tr(th("User"), td(str(whoami()).strip())))
            except Exception:
                rows.append(tr(th("Command Info"), td("Unavailable")))

        sections = [
            h1("Runtime Information"),
            table(tbody(*rows), _id="runtime-info"),
        ]

        if include_environment:
            env_rows = [tr(th(key), td(value)) for key, value in sorted(os.environ.items())]
            sections.append(h2("Environment"))
            sections.append(table(tbody(*env_rows), _id="environment-info"))

        super().__init__(title_text, div(*sections, _id="runtime"), wholepage)


__all__ = [
    "blank_page",
    "maintenance_page",
    "message_page",
    "redirect_page",
    "runtime_page",
    "status_page",
]



"""
domonic.svg
===========

SVG tag constructors and SVG-aware DOM elements for domonic.

The SVG surface is designed to participate in the same tree, event, and
rendering model as the rest of the library rather than living in a separate
mini-framework.
"""

from __future__ import annotations

from typing import Any

from domonic.dom import Element
from domonic.html import a, audio, canvas, iframe, script, style, video

SVG_NAMESPACE = "http://www.w3.org/2000/svg"


class SVGElement(Element):
    """Base SVG element that keeps the SVG namespace on direct constructors."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.namespaceURI = SVG_NAMESPACE


svg_tags = [
    "svg",
    "animate",
    "animateMotion",
    "animateTransform",
    "circle",
    "clipPath",
    "defs",
    "desc",
    "ellipse",
    "filter",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "missing-glyph",
    "mpath",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "tspan",
    "path",
    "rect",
    "stop",
    "switch",
    "symbol",
    "text",
    "textPath",
    "title",
    "use",
    "view",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMorphology",
    "feOffset",
    "feSpecularLighting",
    "feTile",
    "feTurbulence",
    "feDistantLight",
    "fePointLight",
    "feSpotLight",
    "feDropShadow",
    "discard",
    "feDistantLight",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feMergeNode",
    "foreignObject",
    "unknown",
    "cursor",
    "hatchpath",
    "altGlyph",
    "tref",
    "tspan",
    "altGlyphDef",
    "altGlyphItem",
    "glyph",
    "glyphRef",
    "solidcolor",
    "hatch",
    "font",
    "hkern",
    "vkern",
    "animateColor",
    "g",
]


# def svg_init(self, *args, **kwargs):
#     tag.__init__(self, *args, **kwargs)
#     Element.__init__(self, *args, **kwargs)


svg = type("svg", (SVGElement,), {"name": "svg"})

animate = type("animate", (SVGElement,), {"name": "animate"})
animateMotion = type("animateMotion", (SVGElement,), {"name": "animateMotion"})
animateTransform = type("animateTransform", (SVGElement,), {"name": "animateTransform"})
circle = type("circle", (SVGElement,), {"name": "circle"})
clipPath = type("clipPath", (SVGElement,), {"name": "clipPath"})
defs = type("defs", (SVGElement,), {"name": "defs"})
desc = type("desc", (SVGElement,), {"name": "desc"})
ellipse = type("ellipse", (SVGElement,), {"name": "ellipse"})
filter = type("filter", (SVGElement,), {"name": "filter"})
g = type("g", (SVGElement,), {"name": "g"})
image = type("image", (SVGElement,), {"name": "image"})
line = type("line", (SVGElement,), {"name": "line"})
linearGradient = type("linearGradient", (SVGElement,), {"name": "linearGradient"})
marker = type("marker", (SVGElement,), {"name": "marker"})
mask = type("mask", (SVGElement,), {"name": "mask"})
metadata = type("metadata", (SVGElement,), {"name": "metadata"})
mpath = type("mpath", (SVGElement,), {"name": "mpath"})
pattern = type("pattern", (SVGElement,), {"name": "pattern"})
polygon = type("polygon", (SVGElement,), {"name": "polygon"})
polyline = type("polyline", (SVGElement,), {"name": "polyline"})
radialGradient = type("radialGradient", (SVGElement,), {"name": "radialGradient"})

tspan = type("tspan", (SVGElement,), {"name": "tspan"})
path = type("path", (SVGElement,), {"name": "path"})
rect = type("rect", (SVGElement,), {"name": "rect"})
stop = type("stop", (SVGElement,), {"name": "stop"})
switch = type("switch", (SVGElement,), {"name": "switch"})
symbol = type("symbol", (SVGElement,), {"name": "symbol"})
text = type("text", (SVGElement,), {"name": "text"})
textPath = type("textPath", (SVGElement,), {"name": "textPath"})
title = type("title", (SVGElement,), {"name": "title"})
use = type("use", (SVGElement,), {"name": "use"})
view = type("view", (SVGElement,), {"name": "view"})

feBlend = type("feBlend", (SVGElement,), {"name": "feBlend"})
feColorMatrix = type("feColorMatrix", (SVGElement,), {"name": "feColorMatrix"})
feComponentTransfer = type("feComponentTransfer", (SVGElement,), {"name": "feComponentTransfer"})
feComposite = type("feComposite", (SVGElement,), {"name": "feComposite"})
feConvolveMatrix = type("feConvolveMatrix", (SVGElement,), {"name": "feConvolveMatrix"})
feDiffuseLighting = type("feDiffuseLighting", (SVGElement,), {"name": "feDiffuseLighting"})
feDisplacementMap = type("feDisplacementMap", (SVGElement,), {"name": "feDisplacementMap"})
feGaussianBlur = type("feGaussianBlur", (SVGElement,), {"name": "feGaussianBlur"})
feImage = type("feImage", (SVGElement,), {"name": "feImage"})
feMerge = type("feMerge", (SVGElement,), {"name": "feMerge"})
feMorphology = type("feMorphology", (SVGElement,), {"name": "feMorphology"})
feOffset = type("feOffset", (SVGElement,), {"name": "feOffset"})
feSpecularLighting = type("feSpecularLighting", (SVGElement,), {"name": "feSpecularLighting"})
feTile = type("feTile", (SVGElement,), {"name": "feTile"})
feTurbulence = type("feTurbulence", (SVGElement,), {"name": "feTurbulence"})
feDistantLight = type("feDistantLight", (SVGElement,), {"name": "feDistantLight"})
fePointLight = type("fePointLight", (SVGElement,), {"name": "fePointLight"})
feSpotLight = type("feSpotLight", (SVGElement,), {"name": "feSpotLight"})

feDropShadow = type("feDropShadow", (SVGElement,), {"name": "feDropShadow"})
discard = type("discard", (SVGElement,), {"name": "discard"})
feDistantLight = type("feDistantLight", (SVGElement,), {"name": "feDistantLight"})
feFlood = type("feFlood", (SVGElement,), {"name": "feFlood"})
feFuncA = type("feFuncA", (SVGElement,), {"name": "feFuncA"})
feFuncB = type("feFuncB", (SVGElement,), {"name": "feFuncB"})
feFuncG = type("feFuncG", (SVGElement,), {"name": "feFuncG"})
feFuncR = type("feFuncR", (SVGElement,), {"name": "feFuncR"})
feMergeNode = type("feMergeNode", (SVGElement,), {"name": "feMergeNode"})
foreignObject = type("foreignObject", (SVGElement,), {"name": "foreignObject"})
unknown = type("unknown", (SVGElement,), {"name": "unknown"})

cursor = type("cursor", (SVGElement,), {"name": "cursor"})
hatchpath = type("hatchpath", (SVGElement,), {"name": "hatchpath"})
altGlyph = type("altGlyph", (SVGElement,), {"name": "altGlyph"})
tref = type("tref", (SVGElement,), {"name": "tref"})
tspan = type("tspan", (SVGElement,), {"name": "tspan"})
altGlyphDef = type("altGlyphDef", (SVGElement,), {"name": "altGlyphDef"})
altGlyphItem = type("altGlyphItem", (SVGElement,), {"name": "altGlyphItem"})
glyph = type("glyph", (SVGElement,), {"name": "glyph"})
glyphRef = type("glyphRef", (SVGElement,), {"name": "glyphRef"})
solidcolor = type("solidcolor", (SVGElement,), {"name": "solidcolor"})
hatch = type("hatch", (SVGElement,), {"name": "hatch"})
font = type("font", (SVGElement,), {"name": "font"})
hkern = type("hkern", (SVGElement,), {"name": "hkern"})
vkern = type("vkern", (SVGElement,), {"name": "vkern"})
animateColor = type("animateColor", (SVGElement,), {"name": "animateColor"})
missing_glyph = type("missing_glyph", (SVGElement,), {"name": "missing-glyph"})


def create_element(name: str = "custom_svg_tag", *args: Any, **kwargs: Any) -> Element:
    """
    A method for creating SVG tags, including custom or hyphenated ones.
    """
    normalized_name = str(name).strip()
    if not normalized_name:
        normalized_name = "custom_svg_tag"
        name = "custom_svg_tag"

    if normalized_name in svg_tags:
        tag_name = normalized_name.replace("-", "_")
        return globals()[tag_name](*args, **kwargs)

    custom_svg_tag = type("custom_svg_tag", (SVGElement,), {"name": name})
    new_tag = custom_svg_tag(*args, **kwargs)
    new_tag.name = name
    return new_tag


# TODO --
# _set # builtin

# are these obs or new?
# meshgradient 🗑️?
# mesh 🗑️?
# meshpatch 🗑️?
# meshrow 🗑️?

# TODO - just use underscores for these tags?
# color-profile
# font-face
# font-face-format
# font-face-name
# font-face-src
# font-face-uri



"""
    test_svg
    ~~~~~~~~~~~~
    - unit tests for svg
"""

import unittest

from domonic import domonic
from domonic.svg import *


class TestCase(unittest.TestCase):

    # @silence
    def test_domonic_svg(self):
        mysvg = svg()
        assert str(mysvg) == "<svg></svg>"
        mysvg.appendChild(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow")
        )
        assert (
            str(mysvg)
            == '<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'
        )
        mysvg.appendChild(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow")
        )
        assert (
            str(mysvg)
            == '<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'
        )
        # assert mysvg.toxml() == '<?xml version="1.0" encoding="utf-8"?>\n<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="100" height="100" viewBox="0 0 100 100">\n</svg>'

    def test_domonic_cirlce(self):
        test = svg(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"),
            _width="100",
            _height="100",
        )
        # print(test)
        assert (
            str(test)
            == '<svg width="100" height="100"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'
        )

    def test_domonic_node(self):
        circ = svg(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"),
            _width="100",
            _height="100",
        )
        mysvg = svg()
        mysvg.appendChild(circ / 10)
        # print(mysvg)

    def test_svg_namespace_and_factory(self):
        drawing = create_element("svg", _viewBox="0 0 10 10")
        grad = create_element("linearGradient")
        custom = create_element("my-custom-svg")

        self.assertEqual(drawing.namespaceURI, "http://www.w3.org/2000/svg")
        self.assertEqual(grad.namespaceURI, "http://www.w3.org/2000/svg")
        self.assertEqual(str(drawing), '<svg viewBox="0 0 10 10"></svg>')
        self.assertEqual(str(custom), "<my-custom-svg></my-custom-svg>")

    def test_svg_legacy_and_filter_exports(self):
        icon = svg(
            defs(filter(feGaussianBlur(_in="SourceGraphic", _stdDeviation="2"))),
            metadata("info"),
            missing_glyph(),
        )

        self.assertIn("<filter>", str(icon))
        self.assertIn("<metadata>info</metadata>", str(icon))
        self.assertIn("<missing-glyph></missing-glyph>", str(icon))

    def test_svg_tag_exports_real_constructors(self):
        for tag_name in svg_tags:
            python_name = tag_name.replace("-", "_")
            self.assertIn(python_name, globals())
            self.assertEqual(str(globals()[python_name]()), f"<{tag_name}></{tag_name}>")

    # def test_hyphen_elements(self):
    #     test = svg(
    #         missing_glyph(),
    #     )
    #     print(test)

    # def test_font(self):
    #     test = svg(
    #         font_face(),
    #     )
    #     print(test)


if __name__ == "__main__":
    unittest.main()



"""
    test_domonic
    ~~~~~~~~~~~~
    - unit tests for domonic
    # TODO - tests for all bs5 pages
"""

import unittest

from domonic.xml.mathml import *


class TestCase(unittest.TestCase):
    def test_mathml(self):
        somemath = math_(
            maction("x"),
            menclose("x"),
            merror("x"),
            mfenced("x"),
            mfrac("x"),
            mi("x"),
            mmultiscripts("x"),
            mn("x"),
            mo("x"),
            mover("x"),
            mpadded("x"),
            mphantom("x"),
            mroot("x"),
            mrow("x"),
            ms("x"),
            mspace("x"),
            msqrt("x"),
            mstyle("x"),
            msub("x"),
            msubsup("x"),
            msup("x"),
            mtable("x"),
            mtd("x"),
            mtext("x"),
            mtr("x"),
            munder("x"),
            munderover("x"),
            semantics("x"),
            maligngroup("x"),
            malignmark("x"),
            msline("x"),
            msgroup("x"),
            mlongdiv("x"),
            mstyle("x"),
            mprescripts("x"),
            mscarries("x"),
            mscarry("x"),
            munder("x"),
            munderover("x"),
            none("x"),
        )

        print(somemath)


if __name__ == "__main__":
    unittest.main()



"""
    domonic.webapi.url
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/URL

    # TODO - move the unit tests for this class from javascript to webapi
    # TODO - untested

"""
from __future__ import annotations

import urllib.parse
from typing import Any, Callable, Iterable, Iterator

ParamInput = str | dict[str, Any] | Iterable[tuple[str, Any]]


class URL:
    """a-tag extends from URL"""

    def __update__(self) -> None:
        # print( "update URL:", type(self), self  )
        try:
            new = {
                "protocol": self.protocol,
                "hostname": self.hostname,
                "href": self.href,
                "port": self.port,
                "host": self.host,
                "pathname": self.pathname,
                "hash": self.hash,
                "search": self.search,
                "_searchParams": self._searchParams,
            }
            # NOTE - rebuild happening here
            query = new["search"] or ""
            if query and not query.startswith("?"):
                query = "?" + query
            self.url = urllib.parse.urlsplit(
                new["protocol"] + "://" + new["host"] + new["pathname"] + query + new["hash"]
            )

            self.href = self.url.geturl()

        except Exception:  # as e:
            # print('fails on props called by init as they dont exist yet')
            # print(e)
            pass

    def __init__(self, url: str = "", *args: Any, **kwargs: Any) -> None:  # TODO - relative to
        """URL

        builds a url

        Args:
            url (str): a url
        """
        self.url = urllib.parse.urlsplit(url)
        self.href = url  # self.url.geturl()
        self.protocol = self.url.scheme
        self.hostname = self.url.hostname
        self.port = self.url.port
        self.host = self.url.hostname
        self.pathname = self.url.path
        self.hash = ""
        self.search = self.url.query
        self._searchParams = URLSearchParams(self.url.query)

    @property
    def searchParams(self) -> str:
        return self._searchParams.toString()

    @property
    def origin(self) -> str:
        if not self.protocol or not self.host:
            return ""
        return f"{self.protocol}://{self.host}"

    def toString(self) -> str:
        return str(self.href)

    # def toJson

    # @property
    # def href(self):
    # TODO - check js vs tag. does js version remove query?. if so detect self.
    #     return self.href

    # @href.setter
    # def href(self, href:str):
    #     self.url = href
    #     self.href = href

    @property
    def protocol(self) -> str:
        return self.__protocol

    @protocol.setter
    def protocol(self, p: str):
        self.__protocol = p
        # if self.ready : self.__update__() # TODO - this instead of silent err?
        self.__update__()

    @property
    def hostname(self) -> str | None:
        return self.__hostname

    @hostname.setter
    def hostname(self, h: str):
        if h is None:
            return
        if ":" in h:
            h = h.split(":")[0]
        self.__hostname = h
        self.__update__()

    @property
    def port(self) -> int | None:
        return self.__port

    @port.setter
    def port(self, p: int | None):
        self.__port = p
        self.__update__()

    @property
    def host(self) -> str | None:
        if self.hostname is None:
            return None
        if self.port is not None:
            return self.hostname + ":" + str(self.port)
        else:
            return self.hostname

    @host.setter
    def host(self, h: str):
        if h is None:
            return
        p = self.port
        if ":" in h:
            p = int(h.split(":")[1])
            h = h.split(":")[0]
        self.__host = h
        self.hostname = h
        self.port = p
        self.__update__()

    @property
    def pathname(self) -> str:
        return self.__pathname

    @pathname.setter
    def pathname(self, p: str):
        self.__pathname = p
        self.__update__()

    @property
    def search(self) -> str:
        if not self.__search:
            return ""
        return self.__search if self.__search.startswith("?") else "?" + self.__search

    @search.setter
    def search(self, value: str):
        if value is None:
            value = ""
        self.__search = value if value == "" else value.lstrip("?")
        self._searchParams = URLSearchParams(self.__search)
        self.__update__()

    @property
    def hash(self) -> str:
        """ " hash Sets or returns the anchor part (#) of a URL"""
        if "#" in self.href:
            return "#" + self.href.split("#")[1]
        # return ''
        return self.__hash

    @hash.setter
    def hash(self, h: str):
        self.__hash = h
        self.__update__()

        # @property
        # def origin(self):
        """# origin    Returns the protocol, hostname and port number of a URL Location"""

    def __str__(self) -> str:
        return str(self.href)

    # NOTE - node -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # @staticmethod
    # def domainToASCII(domain: str):
    #     """[It returns the Punycode ASCII serialization of the domain.
    #     If domain is an invalid domain, the empty string is returned.]

    #     Args:
    #         domain (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def domainToUnicode(domain: str):
    #     """[returns the Unicode serialization of the domain.
    #     If the domain is invalid, the empty string is returned]

    #     Args:
    #         domain (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def fileURLToPath(url: str):
    #     """[ensures the correct decodings of percent-encoded characters as well as
    #     ensuring a cross-platform valid absolute path string.]

    #     Args:
    #         url (str): [The fully-resolved platform-specific file path.]
    #     """
    #     if url is None:
    #         return
    #     return urllib.parse.unquote(url)

    # @staticmethod
    # def format(URL, options):
    #     """[summary]

    #     Args:
    #         URL ([type]): [description]
    #         options ([type]): [description]
    #     """
    #     pass

    # @staticmethod
    # def pathToFileURL(path: str):
    #     """[summary]

    #     Args:
    #         path (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def urlToHttpOptions(url: str):
    #     """[summary]

    #     Args:
    #         url (str): [description]
    #     """
    #     pass


class URLSearchParams:
    """[utility methods to work with the query string of a URL]"""

    def __init__(self, paramString: ParamInput):  # , **paramsObj):
        """[Returns a URLSearchParams object instance.]

        Args:
            paramString ([type]): [ i.e. q=URLUtils.searchParams&topic=api]
        """
        # TODO - escape
        # import ast
        # TODO - dont think i can do this cant urls params have duplicate keys?
        # self.params = ast.literal_eval(paramString)
        if isinstance(paramString, str):
            if paramString.startswith("?"):
                paramString = paramString[1 : len(paramString)]

            import urllib

            self.params: dict[str, list[str]] = urllib.parse.parse_qs(paramString)
        elif isinstance(paramString, dict):
            self.params = {
                str(key): value if isinstance(value, list) else [str(value)] for key, value in paramString.items()
            }
        elif hasattr(paramString, "__iter__"):
            self.params = {}
            for key, value in paramString:
                self.params.setdefault(str(key), []).append(str(value))
        else:
            raise TypeError(
                f"Malformed paramString.  Must be a string or a dict with dict like items. Got: {paramString}"
            )

    def __iter__(self) -> Iterator[tuple[str, list[str]]]:
        for attr in self.params.items():  # dir(self.params.items()):
            # if not attr.startswith("__"):
            yield attr

    def append(self, key: str, value: str) -> None:
        """Appends a specified key/value pair as a new search parameter"""
        # TODO - ordereddict?
        self.params.setdefault(key, []).append(value)  # [key]=value

    def delete(self, key: str) -> None:
        """Deletes the given search parameter, and its associated value, from the list of all search parameters."""
        del self.params[key]

    def has(self, key: str) -> bool:
        """Returns a Boolean indicating if such a given parameter exists."""
        return key in self.params

    def entries(self) -> Iterable[tuple[str, list[str]]]:
        """Returns an iterator allowing iteration through all key/value pairs contained in this object."""
        return self.params.items()

    def forEach(self, func: Callable[[str, list[str]], Any]) -> None:
        """Allows iteration through all values contained in this object via a callback function."""
        for key, value in self.params.items():
            func(key, value)

    def keys(self) -> Iterable[str]:
        """Returns an iterator allowing iteration through all keys of the key/value pairs contained in this object."""
        return self.params.keys()

    def get(self, key: str) -> str | None:
        """Returns the first value associated with the given search parameter."""
        try:
            return self.params.get(key, None)[0]
        except Exception:
            return None

    def sort(self) -> None:
        """Sorts all key/value pairs, if any, by their keys."""
        self.params = dict(sorted(self.params.items()))

    def values(self) -> Iterable[list[str]]:
        """Returns an iterator allowing iteration through all values of the key/value pairs
        contained in this object."""
        return self.params.values()

    def toString(self) -> str:
        """Returns a string containing a query string suitable for use in a URL."""
        # return '&'.join([str(x) for x in self.params])
        return urllib.parse.urlencode(self.params, doseq=True)
        # return str(self.params)

    def set(self, key: str, value: str) -> None:
        """Sets the value associated with a given search parameter to the given value.
        If there are several values, the others are deleted."""
        self.params[key] = [value]

    def getAll(self, key: str) -> list[str] | None:
        """Returns all the values associated with a given search parameter."""
        return self.params.get(key)

    def __str__(self) -> str:
        return urllib.parse.urlencode(self.params, doseq=True)



"""
    domonic.html
    ====================================

    Generate HTML using python.

"""
from __future__ import annotations

from typing import Any

from domonic.dom import Document  # HTMLOptionsCollection,
from domonic.dom import (Comment, DocumentType, DOMConfig, Element,
                         HTMLAnchorElement, HTMLAreaElement, HTMLAudioElement,
                         HTMLBaseElement, HTMLBaseFontElement, HTMLBodyElement,
                         HTMLBRElement, HTMLButtonElement, HTMLCanvasElement,
                         HTMLContentElement, HTMLDataElement,
                         HTMLDataListElement, HTMLDialogElement,
                         HTMLDivElement, HTMLDListElement, HTMLDocument,
                         HTMLElement, HTMLEmbedElement, HTMLFieldSetElement,
                         HTMLFormControlsCollection, HTMLFormElement,
                         HTMLFrameSetElement, HTMLHeadElement,
                         HTMLHeadingElement, HTMLHRElement, HTMLIFrameElement,
                         HTMLImageElement, HTMLInputElement,
                         HTMLIsIndexElement, HTMLKeygenElement,
                         HTMLLabelElement, HTMLLegendElement, HTMLLIElement,
                         HTMLLinkElement, HTMLMapElement, HTMLMediaElement,
                         HTMLMetaElement, HTMLMeterElement, HTMLModElement,
                         HTMLObjectElement, HTMLOListElement,
                         HTMLOptGroupElement, HTMLOptionElement,
                         HTMLOutputElement, HTMLParagraphElement,
                         HTMLParamElement, HTMLPictureElement, HTMLPreElement,
                         HTMLPortalElement,
                         HTMLProgressElement, HTMLQuoteElement,
                         HTMLScriptElement, HTMLSelectElement,
                         HTMLShadowElement, HTMLSourceElement, HTMLSpanElement,
                         HTMLStyleElement, HTMLTableCaptionElement,
                         HTMLTableCellElement, HTMLTableColElement,
                         HTMLTableDataCellElement, HTMLTableElement,
                         HTMLTableHeaderCellElement, HTMLTableRowElement,
                         HTMLTableSectionElement, HTMLSlotElement,
                         HTMLTemplateElement,
                         HTMLTextAreaElement, HTMLTimeElement,
                         HTMLTitleElement, HTMLTrackElement, HTMLUListElement,
                         HTMLUnknownElement, HTMLVideoElement, Node, Text)
from domonic.webapi.url import URL

html_tags = [
    "figcaption",
    "blockquote",
    "textarea",
    "progress",
    "optgroup",
    "noscript",
    "fieldset",
    "datalist",
    "colgroup",
    "summary",
    "section",
    "details",
    "command",
    "caption",
    "article",
    "address",
    "submit",
    "strong",
    "source",
    "select",
    "script",
    "output",
    "option",
    "legend",
    "keygen",
    "iframe",
    "hgroup",
    "header",
    "footer",
    "figure",
    "canvas",
    "button",
    "video",
    "track",
    "title",
    "thead",
    "tfoot",
    "tbody",
    "table",
    "style",
    "small",
    "param",
    "meter",
    "label",
    "input",
    "audio",
    "aside",
    "applet",
    "object",
    "slot",
    "search",
    "basefont",
    "center",
    "embed",
    "frame",
    "frameset",
    "isindex",
    "listing",
    "menuitem",
    "noframes",
    "plaintext",
    "strike",
    "template",
    "picture",
    "dialog",
    "data",
    "time",
    "span",
    "samp",
    "ruby",
    "meta",
    "menu",
    "mark",
    "link",
    "html",
    "head",
    "form",
    "font",
    "code",
    "cite",
    "body",
    "base",
    "area",
    "abbr",
    "main",
    "dir",
    "mod",
    "wbr",
    "var",
    "sup",
    "sub",
    "nav",
    "map",
    "kbd",
    "ins",
    "img",
    "div",
    "dfn",
    "del",
    "col",
    "bdo",
    "bdi",
    "pre",
    "xmp",
    "ul",
    "tr",
    "th",
    "td",
    "rt",
    "rp",
    "ol",
    "li",
    "hr",
    "h6",
    "h5",
    "h4",
    "h3",
    "h2",
    "h1",
    "em",
    "dt",
    "dl",
    "dd",
    "br",
    "u",
    "s",
    "q",
    "p",
    "i",
    "b",
    "a",
]
# big, blink, bold, tt, var, frameset

_HTML_TAG_LOOKUP = set(html_tags)
_TAG_ALIASES = {"del": "del_"}

html_attributes = [
    "accept",
    "accesskey",
    "loading",
    "action",
    "align",
    "alt",
    "async",
    "autocomplete",
    "autofocus",
    "autoplay",
    "bgcolor",
    "border",
    "charset",
    "checked",
    "cite",
    "class",
    "color",
    "cols",
    "colspan",
    "content",
    "contenteditable",
    "controls",
    "coords",
    "data",
    "datetime",
    "default",
    "defer",
    "dir",
    "dirname",
    "disabled",
    "download",
    "draggable",
    "enctype",
    "for",
    "form",
    "formaction",
    "headers",
    "height",
    "hidden",
    "high",
    "href",
    "hreflang",
    "id",
    "ismap",
    "kind",
    "label",
    "lang",
    "list",
    "loop",
    "low",
    "max",
    "maxlength",
    "media",
    "method",
    "min",
    "multiple",
    "muted",
    "name",
    "novalidate",
    "onabort",
    "onafterprint",
    "onbeforeprint",
    "onbeforeunload",
    "onblur",
    "oncanplay",
    "oncanplaythrough",
    "onchange",
    "onclick",
    "oncontextmenu",
    "oncopy",
    "oncuechange",
    "oncut",
    "ondblclick",
    "ondrag",
    "ondragend",
    "ondragenter",
    "ondragleave",
    "ondragover",
    "ondragstart",
    "ondrop",
    "ondurationchange",
    "onemptied",
    "onended",
    "onerror",
    "onfocus",
    "onhashchange",
    "oninput",
    "oninvalid",
    "onkeydown",
    "onkeypress",
    "onkeyup",
    "onload",
    "onloadeddata",
    "onloadedmetadata",
    "onloadstart",
    "onmousedown",
    "onmousemove",
    "onmouseout",
    "onmouseover",
    "onmouseup",
    "onmousewheel",
    "onoffline",
    "ononline",
    "onpagehide",
    "onpageshow",
    "onpaste",
    "onpause",
    "onplay",
    "onplaying",
    "onpopstate",
    "onprogress",
    "onratechange",
    "onreset",
    "onresize",
    "onscroll",
    "onsearch",
    "onseeked",
    "onseeking",
    "onselect",
    "onstalled",
    "onstorage",
    "onsubmit",
    "onsuspend",
    "ontimeupdate",
    "ontoggle",
    "onunload",
    "onvolumechange",
    "onwaiting",
    "onwheel",
    "open",
    "optimum",
    "pattern",
    "placeholder",
    "poster",
    "preload",
    "readonly",
    "rel",
    "required",
    "reversed",
    "rows",
    "rowspan",
    "sandbox",
    "scope",
    "selected",
    "shape",
    "size",
    "sizes",
    "span",
    "spellcheck",
    "src",
    "srcdoc",
    "srclang",
    "srcset",
    "start",
    "step",
    "style",
    "tabindex",
    "target",
    "title",
    "translate",
    "type",
    "usemap",
    "value",
    "width",
    "wrap",
    "property",
    "integrity",
    "crossorigin",
    "nonce",
    "autocapitalize",
    "enterkeyhint",
    "inputmode",
    "is",
    "itemid",
    "itemprop",
    "itemref",
    "itemscope",
    "itemtype",
    "part",
    "slot",
    "spellcheck",
    "alink",
    "nowrap",
    "vlink",
    "vspace",
    "language",
    "clear",
    "hspace",
    "xmlns",
    "about",
    "allowtransparency",
    "datatype",
    "inlist",
    "prefix",
    "resource",
    "rev",
    "typeof",
    "vocab",  # rdfa
    "playsinline",
    "autopictureinpicture",
    "buffered",
    "controlslist",
    "disableremoteplayback",  # video
]

def render(inp: Node, outp: str = "", to: str | None = None) -> str:
    """
    Render an HTML element or document to a string or file.

    Args:
        inp (obj): A domonic tag. For example div()
        outp (str): An optional output filename
        to (str): An optional output type. if 'pyml' is specified then pyml is returned instead of html.

    Returns:
        str: The rendered HTML string (or PyML if specified).
    
    Examples:
        >>> div()  
        '<div></div>'
        >>> render(div(), outp='output.html')
    """
    if to == "pyml":
        if outp != "":
            with open(outp, "w+") as f:
                f.write(inp.__pyml__())
        return inp.__pyml__()
    # else:
    if outp != "":
        with open(outp, "w+") as f:
            f.write(str(inp))
    return str(inp)


class TemplateError(IndexError):
    def __init__(self, error: Exception, message: str = "TemplateError: "):
        """[raised when a template error occurs]

        Args:
            error ([type]): [the error]
            message (str, optional): [description]. Defaults to "Templating error: ".
        """
        self.error = error
        self.hint = ""
        print(self.error)
        if str(self.error) == "list index out of range":
            self.hint = "MISSING UNDERSCORE ON AN ATTRIBUTE"
        self.message = message + self.hint
        super().__init__(self.message)


tag = Node  # legacy support?. TODO - remove in future? 0.0.9


class closed_tag(Node):
    def __str__(self):
        if DOMConfig.RENDER_OPTIONAL_CLOSING_SLASH:
            if DOMConfig.SPACE_BEFORE_OPTIONAL_CLOSING_SLASH:
                return f"<{self.name}{self.__attributes__} />"
            else:
                return f"<{self.name}{self.__attributes__}/>"
        return f"<{self.name}{self.__attributes__}>"


html = type("html", (HTMLDocument,), {"name": "html"})
body = type("body", (HTMLBodyElement,), {"name": "body"})
head = type("head", (HTMLHeadElement,), {"name": "head"})
script = type("script", (HTMLScriptElement,), {"name": "script"})
style = type("style", (HTMLStyleElement,), {"name": "style"})
h1 = type("h1", (HTMLHeadingElement,), {"name": "h1"})
h2 = type("h2", (HTMLHeadingElement,), {"name": "h2"})
h3 = type("h3", (HTMLHeadingElement,), {"name": "h3"})
h4 = type("h4", (HTMLHeadingElement,), {"name": "h4"})
h5 = type("h5", (HTMLHeadingElement,), {"name": "h5"})
h6 = type("h6", (HTMLHeadingElement,), {"name": "h6"})
p = type("p", (HTMLParagraphElement,), {"name": "p"})

i = type("i", (Element,), {"name": "i"})
b = type("b", (Element,), {"name": "b"})
portal = type("portal", (HTMLPortalElement, Element), {"name": "portal"})


def Atag(self, *args: Any, **kwargs: Any) -> None:
    """
    Base class for the a tag
    """
    # Note - so this only happens on init?
    if kwargs.get("_href", None) is not None:
        URL.__init__(self, url=kwargs["_href"])
        HTMLAnchorElement.__init__(self, *args, **kwargs)

    Element.__init__(self, *args, **kwargs)


a = type("a", (HTMLAnchorElement, Element, URL), {"name": "a", "__init__": Atag})
ul = type("ul", (HTMLUListElement,), {"name": "ul"})
ol = type("ol", (HTMLOListElement,), {"name": "ol"})
li = type("li", (HTMLLIElement,), {"name": "li"})
div = type("div", (HTMLDivElement,), {"name": "div"})

strong = type("strong", (Element,), {"name": "strong"})
blockquote = type("blockquote", (Element,), {"name": "blockquote"})
table = type("table", (HTMLTableElement,), {"name": "table"})
tr = type("tr", (HTMLTableRowElement,), {"name": "tr"})
td = type("td", (HTMLTableCellElement,), {"name": "td"})


class form(HTMLFormElement):

    def __init__(self, *args: Any, **kwargs: Any):
        new_kwargs = {}
        for k, v in kwargs.items():
            if k[0] != "_":
                # print("WARNING: kwarg '{}' should begin with an underscore".format(k))
                new_kwargs[f"_{k}"] = v
            else:
                new_kwargs[k] = v
        kwargs = new_kwargs

        self.name = "form"
        Node.__init__(self, *args, **kwargs)
        Element.__init__(self, *args, **kwargs)

    @property
    def elements(self) -> HTMLFormControlsCollection:
        return super().elements


label = type("label", (HTMLLabelElement,), {"name": "label"})
# label.__doc__ = '''
#                 .. highlight:: python
#                 .. code-block:: python

#                     # used to label form elements. i.e.
#                     label(_for=None, _text=None, **kwargs)
#                     # <label for=""></label>
#                 '''

submit = type("submit", (Element,), {"name": "submit"})
title = type("title", (HTMLTitleElement,), {"name": "title"})
noscript = type("noscript", (Element,), {"name": "noscript"})
section = type("section", (Element,), {"name": "section"})
nav = type("nav", (Element,), {"name": "nav"})
article = type("article", (Element,), {"name": "article"})
aside = type("aside", (Element,), {"name": "aside"})
hgroup = type("hgroup", (Element,), {"name": "hgroup"})
address = type("address", (Element,), {"name": "address"})
pre = type("pre", (HTMLPreElement,), {"name": "pre"})
dl = type("dl", (HTMLDListElement,), {"name": "dl"})
dt = type("dt", (Element,), {"name": "dt"})
dd = type("dd", (Element,), {"name": "dd"})
figure = type("figure", (Element,), {"name": "figure"})
figcaption = type("figcaption", (Element,), {"name": "figcaption"})
em = type("em", (Element,), {"name": "em"})
small = type("small", (Element,), {"name": "small"})
s = type("s", (Element,), {"name": "s"})
cite = type("cite", (Element,), {"name": "cite"})
q = type("q", (HTMLQuoteElement,), {"name": "q"})
dfn = type("dfn", (Element,), {"name": "dfn"})
abbr = type("abbr", (Element,), {"name": "abbr"})
code = type("code", (Element,), {"name": "code"})
var = type("var", (Element,), {"name": "var"})
samp = type("samp", (Element,), {"name": "samp"})
kbd = type("kbd", (Element,), {"name": "kbd"})
sub = type("sub", (Element,), {"name": "sub"})
sup = type("sup", (Element,), {"name": "sup"})
u = type("u", (Element,), {"name": "u"})
mark = type("mark", (Element,), {"name": "mark"})
ruby = type("ruby", (Element,), {"name": "ruby"})
rt = type("rt", (Element,), {"name": "rt"})
rp = type("rp", (Element,), {"name": "rp"})
bdi = type("bdi", (Element,), {"name": "bdi"})
bdo = type("bdo", (Element,), {"name": "bdo"})
span = type("span", (HTMLSpanElement,), {"name": "span"})
ins = type("ins", (Element,), {"name": "ins"})
iframe = type("iframe", (HTMLIFrameElement,), {"name": "iframe"})
video = type("video", (HTMLVideoElement,), {"name": "video"})
audio = type("audio", (HTMLAudioElement,), {"name": "audio"})
canvas = type("canvas", (HTMLCanvasElement,), {"name": "canvas"})
caption = type("caption", (HTMLTableCaptionElement,), {"name": "caption"})
colgroup = type("colgroup", (Element,), {"name": "colgroup"})
tbody = type("tbody", (HTMLTableSectionElement,), {"name": "tbody"})
thead = type("thead", (Element,), {"name": "thead"})  # Note - also should extend HTMLTableSectionElement
tfoot = type("tfoot", (Element,), {"name": "tfoot"})
th = type("th", (HTMLTableHeaderCellElement,), {"name": "th"})
fieldset = type(
    "fieldset", (HTMLFieldSetElement,), {"name": "fieldset"}
)
legend = type("legend", (HTMLLegendElement,), {"name": "legend"})
button = type("button", (HTMLButtonElement,), {"name": "button"})
select = type("select", (HTMLSelectElement,), {"name": "select"})
datalist = type("datalist", (HTMLDataListElement,), {"name": "datalist"})
optgroup = type("optgroup", (HTMLOptGroupElement,), {"name": "optgroup"})
option = type("option", (HTMLOptionElement,), {"name": "option"})
textarea = type("textarea", (HTMLTextAreaElement,), {"name": "textarea"})
output = type("output", (HTMLOutputElement,), {"name": "output"})
progress = type("progress", (HTMLProgressElement,), {"name": "progress"})
meter = type("meter", (HTMLMeterElement,), {"name": "meter"})
details = type("details", (Element,), {"name": "details"})
summary = type("summary", (Element,), {"name": "summary"})
menu = type("menu", (Element,), {"name": "menu"})
menuitem = type("menuitem", (Element,), {"name": "menuitem"})  # dead but may be used
font = type("font", (Element,), {"name": "font"})
header = type("header", (Element,), {"name": "header"})
footer = type("footer", (Element,), {"name": "footer"})
map = type("map", (HTMLMapElement,), {"name": "map"})
object = type("object", (HTMLObjectElement,), {"name": "object"})
del_ = type("del_", (HTMLModElement,), {"name": "del"})
mod = type("mod", (HTMLModElement,), {"name": "mod"})

time = type("time", (HTMLTimeElement,), {"name": "time"})
data = type("data", (HTMLDataElement,), {"name": "data"})

base = type("base", (HTMLBaseElement,), {"name": "base"})
link = type("link", (closed_tag, HTMLLinkElement), {"name": "link"})  # HTMLLinkElement TODO - closed tags
meta = type("meta", (closed_tag, HTMLMetaElement), {"name": "meta"})  # HTMLMetaElement TODO - closed tags
hr = type("hr", (closed_tag, HTMLHRElement), {"name": "hr"})
br = type(
    "br",
    (
        closed_tag,
        HTMLBRElement,
    ),
    {"name": "br"},
)
wbr = type("wbr", (closed_tag, Element), {"name": "wbr"})
img = type("img", (closed_tag, HTMLImageElement), {"name": "img"})
param = type("param", (closed_tag, HTMLParamElement), {"name": "param"})
source = type("source", (closed_tag, HTMLSourceElement), {"name": "source"})
track = type("track", (closed_tag, HTMLTrackElement), {"name": "track"})
area = type("area", (HTMLAreaElement,), {"name": "area"})
col = type("col", (closed_tag, HTMLTableColElement), {"name": "col"})
input = type("input", (closed_tag, HTMLInputElement), {"name": "input"})
keygen = type("keygen", (closed_tag, HTMLKeygenElement), {"name": "keygen"})
command = type("command", (closed_tag, Element), {"name": "command"})

main = type("main", (Element,), {"name": "main"})
slot = type("slot", (HTMLSlotElement,), {"name": "slot"})
search = type("search", (Element,), {"name": "search"})

# obsolete
applet = type("applet", (Element,), {"name": "applet"})
basefont = type("basefont", (HTMLBaseFontElement,), {"name": "basefont"})
center = type("center", (Element,), {"name": "center"})
dir = type("dir", (Element,), {"name": "dir"})
embed = type("embed", (HTMLEmbedElement,), {"name": "embed"})
frame = type("frame", (Element,), {"name": "frame"})
frameset = type("frameset", (HTMLFrameSetElement,), {"name": "frameset"})
isindex = type("isindex", (Element,), {"name": "isindex"})
listing = type("listing", (Element,), {"name": "listing"})
noframes = type("noframes", (Element,), {"name": "noframes"})
plaintext = type("plaintext", (Element,), {"name": "plaintext"})
strike = type("strike", (Element,), {"name": "strike"})
xmp = type("xmp", (Element,), {"name": "xmp"})
# shadow

template = type("template", (HTMLTemplateElement,), {"name": "template"})

picture = type("picture", (HTMLPictureElement,), {"name": "picture"})
dialog = type("dialog", (HTMLDialogElement,), {"name": "dialog"})

# legacy.
doctype = type("doctype", (DocumentType,), {"name": "doctype"})
comment = type("comment", (Comment,), {"name": "comment"})
content = type("content", (HTMLContentElement,), {"name": "content"})


def create_element(name: str = "custom_tag", *args: Any, **kwargs: Any) -> Element:
    """
    A method for creating custom tags

    tag name needs to be set due to custom tags with hyphens can't be classnames.
    i.e. hypenated tags <some-custom-tag></some-custom-tag>
    """
    # checks if already exists
    normalized_name = str(name).strip().lower()
    if not normalized_name:
        normalized_name = "custom_tag"
        name = "custom_tag"
    if normalized_name in _HTML_TAG_LOOKUP:
        tag_name = _TAG_ALIASES.get(normalized_name, normalized_name)
        return globals()[tag_name](*args, **kwargs)

    try:
        from domonic.window import window as domonic_window

        registry = getattr(domonic_window, "customElements", None)
        if registry is not None:
            registered = registry.get(name) or registry.get(normalized_name)
            if registered is not None:
                return registered(*args, **kwargs)
    except Exception:
        pass

    # NOTE: we care calling it custom_tag because it can't have hyphens < Note - use HTMLUnknownElement?
    custom_tag = type("custom_tag", (Element,), {"name": name})
    new_tag = custom_tag(*args, **kwargs)
    new_tag.name = name
    return new_tag



from __future__ import annotations

from typing import ClassVar, Final


class KeyLocation:
    """KeyboardEvent.location constants from the modern UI Events model."""

    STANDARD: ClassVar[int] = 0
    LEFT: ClassVar[int] = 1
    RIGHT: ClassVar[int] = 2
    NUMPAD: ClassVar[int] = 3


class Key:
    """Common KeyboardEvent.key values."""

    ENTER: ClassVar[str] = "Enter"
    TAB: ClassVar[str] = "Tab"
    SPACE: ClassVar[str] = " "
    ESCAPE: ClassVar[str] = "Escape"
    BACKSPACE: ClassVar[str] = "Backspace"
    DELETE: ClassVar[str] = "Delete"
    INSERT: ClassVar[str] = "Insert"
    HOME: ClassVar[str] = "Home"
    END: ClassVar[str] = "End"
    PAGE_UP: ClassVar[str] = "PageUp"
    PAGE_DOWN: ClassVar[str] = "PageDown"
    ARROW_LEFT: ClassVar[str] = "ArrowLeft"
    ARROW_RIGHT: ClassVar[str] = "ArrowRight"
    ARROW_UP: ClassVar[str] = "ArrowUp"
    ARROW_DOWN: ClassVar[str] = "ArrowDown"
    SHIFT: ClassVar[str] = "Shift"
    CONTROL: ClassVar[str] = "Control"
    ALT: ClassVar[str] = "Alt"
    META: ClassVar[str] = "Meta"
    CAPS_LOCK: ClassVar[str] = "CapsLock"
    NUM_LOCK: ClassVar[str] = "NumLock"
    SCROLL_LOCK: ClassVar[str] = "ScrollLock"
    CONTEXT_MENU: ClassVar[str] = "ContextMenu"

    F1: ClassVar[str] = "F1"
    F2: ClassVar[str] = "F2"
    F3: ClassVar[str] = "F3"
    F4: ClassVar[str] = "F4"
    F5: ClassVar[str] = "F5"
    F6: ClassVar[str] = "F6"
    F7: ClassVar[str] = "F7"
    F8: ClassVar[str] = "F8"
    F9: ClassVar[str] = "F9"
    F10: ClassVar[str] = "F10"
    F11: ClassVar[str] = "F11"
    F12: ClassVar[str] = "F12"

    A: ClassVar[str] = "a"
    B: ClassVar[str] = "b"
    C: ClassVar[str] = "c"
    D: ClassVar[str] = "d"
    E: ClassVar[str] = "e"
    F: ClassVar[str] = "f"
    G: ClassVar[str] = "g"
    H: ClassVar[str] = "h"
    I: ClassVar[str] = "i"
    J: ClassVar[str] = "j"
    K: ClassVar[str] = "k"
    L: ClassVar[str] = "l"
    M: ClassVar[str] = "m"
    N: ClassVar[str] = "n"
    O: ClassVar[str] = "o"
    P: ClassVar[str] = "p"
    Q: ClassVar[str] = "q"
    R: ClassVar[str] = "r"
    S: ClassVar[str] = "s"
    T: ClassVar[str] = "t"
    U: ClassVar[str] = "u"
    V: ClassVar[str] = "v"
    W: ClassVar[str] = "w"
    X: ClassVar[str] = "x"
    Y: ClassVar[str] = "y"
    Z: ClassVar[str] = "z"

    DIGIT_0: ClassVar[str] = "0"
    DIGIT_1: ClassVar[str] = "1"
    DIGIT_2: ClassVar[str] = "2"
    DIGIT_3: ClassVar[str] = "3"
    DIGIT_4: ClassVar[str] = "4"
    DIGIT_5: ClassVar[str] = "5"
    DIGIT_6: ClassVar[str] = "6"
    DIGIT_7: ClassVar[str] = "7"
    DIGIT_8: ClassVar[str] = "8"
    DIGIT_9: ClassVar[str] = "9"


class Code:
    """Common KeyboardEvent.code values."""

    ENTER: ClassVar[str] = "Enter"
    TAB: ClassVar[str] = "Tab"
    SPACE: ClassVar[str] = "Space"
    ESCAPE: ClassVar[str] = "Escape"
    BACKSPACE: ClassVar[str] = "Backspace"
    DELETE: ClassVar[str] = "Delete"
    INSERT: ClassVar[str] = "Insert"
    HOME: ClassVar[str] = "Home"
    END: ClassVar[str] = "End"
    PAGE_UP: ClassVar[str] = "PageUp"
    PAGE_DOWN: ClassVar[str] = "PageDown"
    ARROW_LEFT: ClassVar[str] = "ArrowLeft"
    ARROW_RIGHT: ClassVar[str] = "ArrowRight"
    ARROW_UP: ClassVar[str] = "ArrowUp"
    ARROW_DOWN: ClassVar[str] = "ArrowDown"
    SHIFT_LEFT: ClassVar[str] = "ShiftLeft"
    SHIFT_RIGHT: ClassVar[str] = "ShiftRight"
    CONTROL_LEFT: ClassVar[str] = "ControlLeft"
    CONTROL_RIGHT: ClassVar[str] = "ControlRight"
    ALT_LEFT: ClassVar[str] = "AltLeft"
    ALT_RIGHT: ClassVar[str] = "AltRight"
    META_LEFT: ClassVar[str] = "MetaLeft"
    META_RIGHT: ClassVar[str] = "MetaRight"
    CAPS_LOCK: ClassVar[str] = "CapsLock"
    NUM_LOCK: ClassVar[str] = "NumLock"
    SCROLL_LOCK: ClassVar[str] = "ScrollLock"
    CONTEXT_MENU: ClassVar[str] = "ContextMenu"

    F1: ClassVar[str] = "F1"
    F2: ClassVar[str] = "F2"
    F3: ClassVar[str] = "F3"
    F4: ClassVar[str] = "F4"
    F5: ClassVar[str] = "F5"
    F6: ClassVar[str] = "F6"
    F7: ClassVar[str] = "F7"
    F8: ClassVar[str] = "F8"
    F9: ClassVar[str] = "F9"
    F10: ClassVar[str] = "F10"
    F11: ClassVar[str] = "F11"
    F12: ClassVar[str] = "F12"

    KEY_A: ClassVar[str] = "KeyA"
    KEY_B: ClassVar[str] = "KeyB"
    KEY_C: ClassVar[str] = "KeyC"
    KEY_D: ClassVar[str] = "KeyD"
    KEY_E: ClassVar[str] = "KeyE"
    KEY_F: ClassVar[str] = "KeyF"
    KEY_G: ClassVar[str] = "KeyG"
    KEY_H: ClassVar[str] = "KeyH"
    KEY_I: ClassVar[str] = "KeyI"
    KEY_J: ClassVar[str] = "KeyJ"
    KEY_K: ClassVar[str] = "KeyK"
    KEY_L: ClassVar[str] = "KeyL"
    KEY_M: ClassVar[str] = "KeyM"
    KEY_N: ClassVar[str] = "KeyN"
    KEY_O: ClassVar[str] = "KeyO"
    KEY_P: ClassVar[str] = "KeyP"
    KEY_Q: ClassVar[str] = "KeyQ"
    KEY_R: ClassVar[str] = "KeyR"
    KEY_S: ClassVar[str] = "KeyS"
    KEY_T: ClassVar[str] = "KeyT"
    KEY_U: ClassVar[str] = "KeyU"
    KEY_V: ClassVar[str] = "KeyV"
    KEY_W: ClassVar[str] = "KeyW"
    KEY_X: ClassVar[str] = "KeyX"
    KEY_Y: ClassVar[str] = "KeyY"
    KEY_Z: ClassVar[str] = "KeyZ"

    DIGIT_0: ClassVar[str] = "Digit0"
    DIGIT_1: ClassVar[str] = "Digit1"
    DIGIT_2: ClassVar[str] = "Digit2"
    DIGIT_3: ClassVar[str] = "Digit3"
    DIGIT_4: ClassVar[str] = "Digit4"
    DIGIT_5: ClassVar[str] = "Digit5"
    DIGIT_6: ClassVar[str] = "Digit6"
    DIGIT_7: ClassVar[str] = "Digit7"
    DIGIT_8: ClassVar[str] = "Digit8"
    DIGIT_9: ClassVar[str] = "Digit9"

    NUMPAD_0: ClassVar[str] = "Numpad0"
    NUMPAD_1: ClassVar[str] = "Numpad1"
    NUMPAD_2: ClassVar[str] = "Numpad2"
    NUMPAD_3: ClassVar[str] = "Numpad3"
    NUMPAD_4: ClassVar[str] = "Numpad4"
    NUMPAD_5: ClassVar[str] = "Numpad5"
    NUMPAD_6: ClassVar[str] = "Numpad6"
    NUMPAD_7: ClassVar[str] = "Numpad7"
    NUMPAD_8: ClassVar[str] = "Numpad8"
    NUMPAD_9: ClassVar[str] = "Numpad9"
    NUMPAD_ADD: ClassVar[str] = "NumpadAdd"
    NUMPAD_DECIMAL: ClassVar[str] = "NumpadDecimal"
    NUMPAD_DIVIDE: ClassVar[str] = "NumpadDivide"
    NUMPAD_ENTER: ClassVar[str] = "NumpadEnter"
    NUMPAD_MULTIPLY: ClassVar[str] = "NumpadMultiply"
    NUMPAD_SUBTRACT: ClassVar[str] = "NumpadSubtract"


_LEGACY_KEYCODE_TO_KEY: Final[dict[str, str]] = {
    "8": Key.BACKSPACE,
    "9": Key.TAB,
    "13": Key.ENTER,
    "16": Key.SHIFT,
    "17": Key.CONTROL,
    "18": Key.ALT,
    "20": Key.CAPS_LOCK,
    "27": Key.ESCAPE,
    "32": Key.SPACE,
    "33": Key.PAGE_UP,
    "34": Key.PAGE_DOWN,
    "35": Key.END,
    "36": Key.HOME,
    "37": Key.ARROW_LEFT,
    "38": Key.ARROW_UP,
    "39": Key.ARROW_RIGHT,
    "40": Key.ARROW_DOWN,
    "45": Key.INSERT,
    "46": Key.DELETE,
    "91": Key.META,
    "93": Key.CONTEXT_MENU,
    "144": Key.NUM_LOCK,
    "145": Key.SCROLL_LOCK,
}
_LEGACY_KEYCODE_TO_KEY.update({str(48 + i): str(i) for i in range(10)})
_LEGACY_KEYCODE_TO_KEY.update({str(65 + i): chr(97 + i) for i in range(26)})
_LEGACY_KEYCODE_TO_KEY.update({str(111 + i): f"F{i}" for i in range(1, 13)})

_KEY_TO_CODE: Final[dict[str, str]] = {
    Key.ENTER: Code.ENTER,
    Key.TAB: Code.TAB,
    Key.SPACE: Code.SPACE,
    Key.ESCAPE: Code.ESCAPE,
    Key.BACKSPACE: Code.BACKSPACE,
    Key.DELETE: Code.DELETE,
    Key.INSERT: Code.INSERT,
    Key.HOME: Code.HOME,
    Key.END: Code.END,
    Key.PAGE_UP: Code.PAGE_UP,
    Key.PAGE_DOWN: Code.PAGE_DOWN,
    Key.ARROW_LEFT: Code.ARROW_LEFT,
    Key.ARROW_RIGHT: Code.ARROW_RIGHT,
    Key.ARROW_UP: Code.ARROW_UP,
    Key.ARROW_DOWN: Code.ARROW_DOWN,
    Key.SHIFT: Code.SHIFT_LEFT,
    Key.CONTROL: Code.CONTROL_LEFT,
    Key.ALT: Code.ALT_LEFT,
    Key.META: Code.META_LEFT,
    Key.CAPS_LOCK: Code.CAPS_LOCK,
    Key.NUM_LOCK: Code.NUM_LOCK,
    Key.SCROLL_LOCK: Code.SCROLL_LOCK,
    Key.CONTEXT_MENU: Code.CONTEXT_MENU,
}
_KEY_TO_CODE.update({chr(97 + i): f"Key{chr(65 + i)}" for i in range(26)})
_KEY_TO_CODE.update({str(i): f"Digit{i}" for i in range(10)})
_KEY_TO_CODE.update({f"F{i}": f"F{i}" for i in range(1, 13)})

_KEY_TO_LEGACY_KEYCODE: Final[dict[str, int]] = {
    key: int(code) for code, key in _LEGACY_KEYCODE_TO_KEY.items()
}


class KeyCode:
    """Legacy KeyboardEvent.keyCode constants kept for compatibility."""

    A: ClassVar[str] = "65"
    B: ClassVar[str] = "66"
    C: ClassVar[str] = "67"
    D: ClassVar[str] = "68"
    E: ClassVar[str] = "69"
    F: ClassVar[str] = "70"
    G: ClassVar[str] = "71"
    H: ClassVar[str] = "72"
    I: ClassVar[str] = "73"
    J: ClassVar[str] = "74"
    K: ClassVar[str] = "75"
    L: ClassVar[str] = "76"
    M: ClassVar[str] = "77"
    N: ClassVar[str] = "78"
    O: ClassVar[str] = "79"
    P: ClassVar[str] = "80"
    Q: ClassVar[str] = "81"
    R: ClassVar[str] = "82"
    S: ClassVar[str] = "83"
    T: ClassVar[str] = "84"
    U: ClassVar[str] = "85"
    V: ClassVar[str] = "86"
    W: ClassVar[str] = "87"
    X: ClassVar[str] = "88"
    Y: ClassVar[str] = "89"
    Z: ClassVar[str] = "90"

    NUMBER_0: ClassVar[str] = "48"
    NUMBER_1: ClassVar[str] = "49"
    NUMBER_2: ClassVar[str] = "50"
    NUMBER_3: ClassVar[str] = "51"
    NUMBER_4: ClassVar[str] = "52"
    NUMBER_5: ClassVar[str] = "53"
    NUMBER_6: ClassVar[str] = "54"
    NUMBER_7: ClassVar[str] = "55"
    NUMBER_8: ClassVar[str] = "56"
    NUMBER_9: ClassVar[str] = "57"

    NUMPAD_0: ClassVar[str] = "96"
    NUMPAD_1: ClassVar[str] = "97"
    NUMPAD_2: ClassVar[str] = "98"
    NUMPAD_3: ClassVar[str] = "99"
    NUMPAD_4: ClassVar[str] = "100"
    NUMPAD_5: ClassVar[str] = "101"
    NUMPAD_6: ClassVar[str] = "102"
    NUMPAD_7: ClassVar[str] = "103"
    NUMPAD_8: ClassVar[str] = "104"
    NUMPAD_9: ClassVar[str] = "105"
    NUMPAD_MULTIPLY: ClassVar[str] = "106"
    NUMPAD_ADD: ClassVar[str] = "107"
    NUMPAD_ENTER: ClassVar[str] = "108"
    NUMPAD_SUBTRACT: ClassVar[str] = "109"
    NUMPAD_DECIMAL: ClassVar[str] = "110"
    NUMPAD_DIVIDE: ClassVar[str] = "111"

    F1: ClassVar[str] = "112"
    F2: ClassVar[str] = "113"
    F3: ClassVar[str] = "114"
    F4: ClassVar[str] = "115"
    F5: ClassVar[str] = "116"
    F6: ClassVar[str] = "117"
    F7: ClassVar[str] = "118"
    F8: ClassVar[str] = "119"
    F9: ClassVar[str] = "120"
    F10: ClassVar[str] = "121"
    F11: ClassVar[str] = "122"
    F12: ClassVar[str] = "123"

    BACKSPACE: ClassVar[str] = "8"
    TAB: ClassVar[str] = "9"
    ENTER: ClassVar[str] = "13"
    SHIFT: ClassVar[str] = "16"
    CONTROL: ClassVar[str] = "17"
    ALT: ClassVar[str] = "18"
    CAPS_LOCK: ClassVar[str] = "20"
    ESCAPE: ClassVar[str] = "27"
    SPACE: ClassVar[str] = "32"
    PAGE_UP: ClassVar[str] = "33"
    PAGE_DOWN: ClassVar[str] = "34"
    END: ClassVar[str] = "35"
    HOME: ClassVar[str] = "36"
    LEFT: ClassVar[str] = "37"
    UP: ClassVar[str] = "38"
    RIGHT: ClassVar[str] = "39"
    DOWN: ClassVar[str] = "40"
    INSERT: ClassVar[str] = "45"
    DELETE: ClassVar[str] = "46"
    LEFT_COMMAND: ClassVar[str] = "91"
    RIGHT_COMMAND: ClassVar[str] = "93"
    WINDOWS: ClassVar[str] = "91"
    MENU: ClassVar[str] = "93"
    NUM_LOCK: ClassVar[str] = "144"
    SCROLL_LOCK: ClassVar[str] = "145"

    MODIFIER_KEYS: ClassVar[frozenset[str]] = frozenset(
        {SHIFT, CONTROL, ALT, LEFT_COMMAND, RIGHT_COMMAND, CAPS_LOCK, NUM_LOCK, SCROLL_LOCK}
    )

    @classmethod
    def is_modifier(cls, key_code: str) -> bool:
        return key_code in cls.MODIFIER_KEYS

    @classmethod
    def to_key(cls, key_code: str | int) -> str | None:
        return _LEGACY_KEYCODE_TO_KEY.get(str(key_code))

    @classmethod
    def to_code(cls, key_code: str | int) -> str | None:
        key = cls.to_key(key_code)
        return _KEY_TO_CODE.get(key) if key is not None else None

    @classmethod
    def from_key(cls, key: str) -> int | None:
        return _KEY_TO_LEGACY_KEYCODE.get(normalize_key(key))


def normalize_key(key: str | None) -> str:
    """Normalize browser-ish key strings to the modern KeyboardEvent.key form."""
    if key in (None, ""):
        return ""
    aliases = {
        "Esc": Key.ESCAPE,
        "Spacebar": Key.SPACE,
        "Left": Key.ARROW_LEFT,
        "Right": Key.ARROW_RIGHT,
        "Up": Key.ARROW_UP,
        "Down": Key.ARROW_DOWN,
        "Del": Key.DELETE,
        "OS": Key.META,
    }
    normalized = aliases.get(key, key)
    if len(normalized) == 1 and normalized.isalpha():
        return normalized.lower()
    return normalized


def normalize_code(code: str | None) -> str:
    """Normalize shorthand physical key names to KeyboardEvent.code values."""
    if code in (None, ""):
        return ""
    if len(code) == 1 and code.isalpha():
        return f"Key{code.upper()}"
    if len(code) == 1 and code.isdigit():
        return f"Digit{code}"
    return code


__all__ = ["Code", "Key", "KeyCode", "KeyLocation", "normalize_code", "normalize_key"]



"""
    test_collada
    ~~~~~~~~~~~~
"""

import unittest

from domonic import domonic
from domonic.decorators import silence
from domonic.dom import *
from domonic.html import *
from domonic.xml.x3d import *


class TestCase(unittest.TestCase):

    # @silence
    def test_domonic_x3dom(self):
        x3dom_test = html(
            head(
                meta(**{"_http-equiv": "X-UA-Compatible"}, _content="IE=edge"),
                title("My first X3DOM page"),
                script(_type="text/javascript", _src="https://www.x3dom.org/download/x3dom.js"),
                script(
                    """
				$(document).ready(function(){
				    var screenshotCount = 0;

				    //Every time the user clicks on the 'take screenhot' button
				    $("#btnTakeScreenshot").on("click", function() {
				        //Get data url from the runtime
				        var imgUrl = document.getElementById("canvas").runtime.getScreenshot();

				        //Create preview image...
				        var newScreenshotImage = document.createElement('img');
				        newScreenshotImage.src = imgUrl;
				        newScreenshotImage.id = "screenshot_" + screenshotCount;
				        $('#screenshotPreviews').append(newScreenshotImage);

				        //...and download link
				        var newScreenshotDownloadLink = document.createElement('a');
				        newScreenshotDownloadLink.href = imgUrl;
				        newScreenshotDownloadLink.download = "screenshot_" + screenshotCount + ".png";
				        newScreenshotDownloadLink.innerHTML = "Speichern";
				        $('#screenshotPreviews').append(newScreenshotDownloadLink);

				        screenshotCount++;
				        $('#screenshotCount').html(screenshotCount);
				    });
				});
		    	"""
                ),
                link(_rel="stylesheet", _type="text/css", _href="https://www.x3dom.org/download/x3dom.css"),
            ),
            body(
                h1("Animate Objects with X3DOM!"),
                p("Learn how to animate objects."),
                x3d(_width="500px", _height="400px").append(
                    scene(
                        transform(_DEF="ball").append(shape(appearance(material(_diffuseColor="1 0 0")), sphere())),
                        timeSensor(_DEF="time", _cycleInterval="2", _loop="true"),
                        PositionInterpolator(_DEF="move", _key="0 0.5 1", _keyValue="0 0 0  0 3 0  0 0 0"),
                        Route(_fromNode="time", _fromField="fraction_changed", _toNode="move", _toField="set_fraction"),
                        Route(_fromNode="move", _fromField="value_changed", _toNode="ball", _toField="translation"),
                    )
                ),
            ),
        )

        render(str(x3dom_test))  # , "sphere_test.html" )


# def test_aframe(self):
# 		pass


if __name__ == "__main__":
    unittest.main()



"""
    domonic.webapi.clipboard
    ====================================

    TODO - make this clipboard same as the webapi and make it work for linux, mac and windows

    https://developer.mozilla.org/en-US/docs/Web/API/Clipboard
    
"""

import sys
mac = sys.platform == "darwin"
windows = sys.platform == "win32"
linux = sys.platform == "linux"

try:
    import pyperclip
except ImportError:  # pragma: no cover - optional dependency
    pyperclip = None


_clipboard_fallback = ""


class Clipboard:
    """
    The Clipboard API provides the ability to respond to clipboard commands (cut, copy, and paste) as well as to asynchronously read from and write to the system clipboard.
    """

    def __init__(self):
        pass

    def writeText(self, data):
        """
        Writes the given text to the clipboard.
        """
        global _clipboard_fallback
        _clipboard_fallback = data
        if pyperclip is not None:
            pyperclip.copy(data)
        return data

    def readText(self):
        if pyperclip is not None:
            return pyperclip.paste()
        return _clipboard_fallback

    def write(self, data):
        return self.writeText(data)

    def read(self):
        return self.readText()

    def writeHTML(self, data):
        pass

    def readHTML(self):
        pass

    def writeImage(self, data):
        pass

    def readImage(self):
        pass

    def writeBuffer(self, data):
        pass

    def readBuffer(self):
        pass

    def writeData(self, data):
        return self.writeText(data)

    def readData(self):
        return self.readText()



# class ClipboardData:
#     def __init__(self):
#         self.data = None



import io
import sys
import tempfile
import unittest
from argparse import Namespace
from contextlib import redirect_stdout
from unittest.mock import MagicMock, patch

from domonic.__main__ import do_things, project


class TestCLI(unittest.TestCase):
    def _base_args(self, **overrides):
        args = Namespace(
            help=False,
            version=False,
            project=None,
            eval=None,
            assets=False,
            download=None,
            xpath=None,
            query=None,
            xpath_file=None,
            query_file=None,
            xpath_stdin=None,
            query_stdin=None,
            text=False,
            attr=None,
            count=False,
            first=False,
            html2pyml=None,
            server=None,
        )
        for key, value in overrides.items():
            setattr(args, key, value)
        return args

    def test_xpath_cli_outputs_nodes(self):
        mocked_nodes = ["<a href='https://example.com'>Example</a>"]
        mocked_result = MagicMock(nodes=mocked_nodes)
        mocked_expression = MagicMock()
        mocked_expression.evaluate.return_value = mocked_result
        mocked_evaluator = MagicMock()
        mocked_evaluator.createExpression.return_value = mocked_expression
        fake_requests = MagicMock()
        fake_requests.get.return_value.text = "<html><body><a href='https://example.com'>Example</a></body></html>"

        with (
            patch.dict(sys.modules, {"requests": fake_requests}),
            patch("domonic.domonic.parseString", return_value=object()) as mock_parse,
            patch("domonic.webapi.xpath.XPathEvaluator", return_value=mocked_evaluator),
            redirect_stdout(io.StringIO()) as stdout,
        ):
            do_things(self._base_args(xpath=["https://example.com", "//a"]))

        fake_requests.get.assert_called_once_with("https://example.com")
        mock_parse.assert_called_once()
        mocked_evaluator.createExpression.assert_called_once_with("//a")
        self.assertIn("Example", stdout.getvalue())

    def test_query_cli_outputs_matching_nodes(self):
        fake_document = MagicMock()
        fake_document.querySelectorAll.return_value = ["<a class='cta'>Call to action</a>"]

        with (
            patch("domonic.window.window") as mock_window,
            redirect_stdout(io.StringIO()) as stdout,
        ):
            mock_window.document = fake_document

            do_things(self._base_args(query=["https://example.com", ".cta"]))

        self.assertEqual(mock_window.location, "https://example.com")
        fake_document.querySelectorAll.assert_called_once_with(".cta")
        self.assertIn("Call to action", stdout.getvalue())

    def test_query_file_cli_supports_attr_and_first(self):
        fake_node = MagicMock()
        fake_node.href = "https://example.com/docs"
        fake_page = MagicMock()
        fake_page.querySelectorAll.return_value = [fake_node, MagicMock(href="https://example.com/ignored")]

        with tempfile.NamedTemporaryFile("w", suffix=".html", delete=False) as handle:
            handle.write("<html><body><a href='https://example.com/docs'>Docs</a></body></html>")
            file_path = handle.name

        try:
            with (
                patch("domonic.domonic.parseString", return_value=fake_page),
                redirect_stdout(io.StringIO()) as stdout,
            ):
                do_things(
                    self._base_args(
                        query_file=[file_path, "a"],
                        attr="href",
                        first=True,
                    )
                )
        finally:
            import os

            os.unlink(file_path)

        fake_page.querySelectorAll.assert_called_once_with("a")
        self.assertEqual(stdout.getvalue().strip(), "https://example.com/docs")

    def test_xpath_stdin_cli_supports_count(self):
        mocked_result = MagicMock(nodes=["<a>One</a>", "<a>Two</a>"])
        mocked_expression = MagicMock()
        mocked_expression.evaluate.return_value = mocked_result
        mocked_evaluator = MagicMock()
        mocked_evaluator.createExpression.return_value = mocked_expression

        with (
            patch("domonic.domonic.parseString", return_value=object()),
            patch("domonic.webapi.xpath.XPathEvaluator", return_value=mocked_evaluator),
            patch("sys.stdin", io.StringIO("<html><body><a>One</a><a>Two</a></body></html>")),
            redirect_stdout(io.StringIO()) as stdout,
        ):
            do_things(self._base_args(xpath_stdin="//a", count=True))

        self.assertEqual(stdout.getvalue().strip(), "2")

    def test_xpath_cli_uses_piped_stdin_when_only_expression_is_passed(self):
        mocked_result = MagicMock(nodes=["<title>Example</title>"])
        mocked_expression = MagicMock()
        mocked_expression.evaluate.return_value = mocked_result
        mocked_evaluator = MagicMock()
        mocked_evaluator.createExpression.return_value = mocked_expression

        with (
            patch("domonic.domonic.parseString", return_value=object()) as mock_parse,
            patch("domonic.webapi.xpath.XPathEvaluator", return_value=mocked_evaluator),
            patch("sys.stdin", io.StringIO("<html><head><title>Example</title></head></html>")),
            patch("sys.stdin.isatty", return_value=False),
            redirect_stdout(io.StringIO()) as stdout,
        ):
            do_things(self._base_args(xpath=["//title"]))

        mock_parse.assert_called_once()
        mocked_evaluator.createExpression.assert_called_once_with("//title")
        self.assertIn("Example", stdout.getvalue())

    def test_query_cli_uses_piped_stdin_when_only_selector_is_passed(self):
        fake_page = MagicMock()
        fake_page.querySelectorAll.return_value = ["<a class='cta'>CTA</a>"]

        with (
            patch("domonic.domonic.parseString", return_value=fake_page),
            patch("sys.stdin", io.StringIO("<html><body><a class='cta'>CTA</a></body></html>")),
            patch("sys.stdin.isatty", return_value=False),
            redirect_stdout(io.StringIO()) as stdout,
        ):
            do_things(self._base_args(query=["a.cta"]))

        fake_page.querySelectorAll.assert_called_once_with("a.cta")
        self.assertIn("CTA", stdout.getvalue())

    def test_xpath_requires_url_and_expression(self):
        with patch("sys.stdin.isatty", return_value=True):
            with self.assertRaisesRegex(ValueError, "xpath expects exactly 2 arguments"):
                do_things(self._base_args(xpath=["https://example.com"]))

    def test_query_requires_url_and_selector(self):
        with patch("sys.stdin.isatty", return_value=True):
            with self.assertRaisesRegex(ValueError, "query expects exactly 2 arguments"):
                do_things(self._base_args(query=["https://example.com"]))

    def test_project_rejects_unknown_server_choice(self):
        with self.assertRaisesRegex(ValueError, "Unsupported server"):
            project("tmp-project", server_choice="not-a-real-server")


if __name__ == "__main__":
    unittest.main()



"""
    domonic.x3d
    ====================================
    Generate x3d with python 3

"""

from domonic.dom import Element
from domonic.html import closed_tag

x3d = X3D = type("x3d", (Element,), {"name": "x3d"})
scene = Scene = type("scene", (Element,), {"name": "scene"})
material = Material = type("material", (Element,), {"name": "material"})
appearance = Appearance = type("appearance", (Element,), {"name": "appearance"})
sphere = Sphere = type("sphere", (Element,), {"name": "sphere"})
shape = Shape = type("shape", (Element,), {"name": "shape"})
transform = Transform = type("transform", (Element,), {"name": "transform"})
timeSensor = TimeSensor = type("timeSensor", (Element,), {"name": "timeSensor"})
inline = Inline = type("inline", (Element,), {"name": "inline"})
box = Box = type("box", (Element,), {"name": "box"})
plane = Plane = type("plane", (Element,), {"name": "plane"})

# TODO - go through examples to find which are usually lower. i.e before they allowed mised cases
positionInterpolator = PositionInterpolator = type("PositionInterpolator", (Element,), {"name": "PositionInterpolator"})
route = Route = type("Route", (Element,), {"name": "Route"})
anchor = Anchor = type("Anchor", (Element,), {"name": "Anchor"})
arc2D = Arc2D = type("Arc2D", (Element,), {"name": "Arc2D"})
arcClose2D = ArcClose2D = type("ArcClose2D", (Element,), {"name": "ArcClose2D"})
audioClip = AudioClip = type("AudioClip", (Element,), {"name": "AudioClip"})
background = Background = type("Background", (Element,), {"name": "Background"})
ballJoint = BallJoint = type("BallJoint", (Element,), {"name": "BallJoint"})
billboard = Billboard = type("Billboard", (Element,), {"name": "Billboard"})
binaryGeometry = BinaryGeometry = type("BinaryGeometry", (Element,), {"name": "BinaryGeometry"})
blendedVolumeStyle = BlendedVolumeStyle = type("BlendedVolumeStyle", (Element,), {"name": "BlendedVolumeStyle"})
blendMode = BlendMode = type("BlendMode", (Element,), {"name": "BlendMode"})
block = Block = type("Block", (Element,), {"name": "Block"})
boundaryEnhancementVolumeStyle = BoundaryEnhancementVolumeStyle = type(
    "BoundaryEnhancementVolumeStyle",
    (Element,),
    {"name": "BoundaryEnhancementVolumeStyle"},
)
bufferAccessor = BufferAccessor = type("BufferAccessor", (Element,), {"name": "BufferAccessor"})
bufferGeometry = BufferGeometry = type("BufferGeometry", (Element,), {"name": "BufferGeometry"})
bufferView = BufferView = type("BufferView", (Element,), {"name": "BufferView"})
cADAssembly = CADAssembly = type("CADAssembly", (Element,), {"name": "CADAssembly"})
cADFace = CADFace = type("CADFace", (Element,), {"name": "CADFace"})
cADLayer = CADLayer = type("CADLayer", (Element,), {"name": "CADLayer"})
cADPart = CADPart = type("CADPart", (Element,), {"name": "CADPart"})
cartoonVolumeStyle = CartoonVolumeStyle = type("CartoonVolumeStyle", (Element,), {"name": "CartoonVolumeStyle"})
circle2D = Circle2D = type("Circle2D", (Element,), {"name": "Circle2D"})
clipPlane = ClipPlane = type("ClipPlane", (Element,), {"name": "ClipPlane"})
collidableShape = CollidableShape = type("CollidableShape", (Element,), {"name": "CollidableShape"})
collision = Collision = type("Collision", (Element,), {"name": "Collision"})
collisionCollection = CollisionCollection = type("CollisionCollection", (Element,), {"name": "CollisionCollection"})
collisionSensor = CollisionSensor = type("CollisionSensor", (Element,), {"name": "CollisionSensor"})
color = Color = type("Color", (Element,), {"name": "Color"})
colorChaser = ColorChaser = type("ColorChaser", (Element,), {"name": "ColorChaser"})
colorDamper = ColorDamper = type("ColorDamper", (Element,), {"name": "ColorDamper"})
colorInterpolator = ColorInterpolator = type("ColorInterpolator", (Element,), {"name": "ColorInterpolator"})
colorMaskMode = ColorMaskMode = type("ColorMaskMode", (Element,), {"name": "ColorMaskMode"})
colorRGBA = ColorRGBA = type("ColorRGBA", (Element,), {"name": "ColorRGBA"})
commonSurfaceShader = CommonSurfaceShader = type("CommonSurfaceShader", (Element,), {"name": "CommonSurfaceShader"})
composedCubeMapTexture = ComposedCubeMapTexture = type(
    "ComposedCubeMapTexture", (Element,), {"name": "ComposedCubeMapTexture"}
)
composedShader = ComposedShader = type("ComposedShader", (Element,), {"name": "ComposedShader"})
composedTexture3D = ComposedTexture3D = type("ComposedTexture3D", (Element,), {"name": "ComposedTexture3D"})
composedVolumeStyle = ComposedVolumeStyle = type("ComposedVolumeStyle", (Element,), {"name": "ComposedVolumeStyle"})
cone = Cone = type("Cone", (Element,), {"name": "Cone"})
coordinate = Coordinate = type("Coordinate", (Element,), {"name": "Coordinate"})
coordinateDamper = CoordinateDamper = type("CoordinateDamper", (Element,), {"name": "CoordinateDamper"})
coordinateDouble = CoordinateDouble = type("CoordinateDouble", (Element,), {"name": "CoordinateDouble"})
coordinateInterpolator = CoordinateInterpolator = type(
    "CoordinateInterpolator", (Element,), {"name": "CoordinateInterpolator"}
)
cylinder = Cylinder = type("Cylinder", (Element,), {"name": "Cylinder"})
cylinderSensor = CylinderSensor = type("CylinderSensor", (Element,), {"name": "CylinderSensor"})
depthMode = DepthMode = type("DepthMode", (Element,), {"name": "DepthMode"})
directionalLight = DirectionalLight = type("DirectionalLight", (Element,), {"name": "DirectionalLight"})
dish = Dish = type("Dish", (Element,), {"name": "Dish"})
disk2D = Disk2D = type("Disk2D", (Element,), {"name": "Disk2D"})
doubleAxisHingeJoint = DoubleAxisHingeJoint = type("DoubleAxisHingeJoint", (Element,), {"name": "DoubleAxisHingeJoint"})
dynamicLOD = DynamicLOD = type("DynamicLOD", (Element,), {"name": "DynamicLOD"})
edgeEnhancementVolumeStyle = EdgeEnhancementVolumeStyle = type(
    "EdgeEnhancementVolumeStyle", (Element,), {"name": "EdgeEnhancementVolumeStyle"}
)
elevationGrid = ElevationGrid = type("ElevationGrid", (Element,), {"name": "ElevationGrid"})
environment = Environment = type("Environment", (Element,), {"name": "Environment"})
extrusion = Extrusion = type("Extrusion", (Element,), {"name": "Extrusion"})
field = Field = type("Field", (Element,), {"name": "Field"})
floatVertexAttribute = FloatVertexAttribute = type("FloatVertexAttribute", (Element,), {"name": "FloatVertexAttribute"})
fog = Fog = type("Fog", (Element,), {"name": "Fog"})
fontStyle = FontStyle = type("FontStyle", (Element,), {"name": "FontStyle"})
generatedCubeMapTexture = GeneratedCubeMapTexture = type(
    "GeneratedCubeMapTexture", (Element,), {"name": "GeneratedCubeMapTexture"}
)
geoCoordinate = GeoCoordinate = type("GeoCoordinate", (Element,), {"name": "GeoCoordinate"})
geoElevationGrid = GeoElevationGrid = type("GeoElevationGrid", (Element,), {"name": "GeoElevationGrid"})
geoLocation = GeoLocation = type("GeoLocation", (Element,), {"name": "GeoLocation"})
geoLOD = GeoLOD = type("GeoLOD", (Element,), {"name": "GeoLOD"})
geoMetadata = GeoMetadata = type("GeoMetadata", (Element,), {"name": "GeoMetadata"})
geoOrigin = GeoOrigin = type("GeoOrigin", (Element,), {"name": "GeoOrigin"})
geoPositionInterpolator = GeoPositionInterpolator = type(
    "GeoPositionInterpolator", (Element,), {"name": "GeoPositionInterpolator"}
)
geoTransform = GeoTransform = type("GeoTransform", (Element,), {"name": "GeoTransform"})
geoViewpoint = GeoViewpoint = type("GeoViewpoint", (Element,), {"name": "GeoViewpoint"})
group = Group = type("Group", (Element,), {"name": "Group"})
hAnimDisplacer = HAnimDisplacer = type("HAnimDisplacer", (Element,), {"name": "HAnimDisplacer"})
hAnimHumanoid = HAnimHumanoid = type("HAnimHumanoid", (Element,), {"name": "HAnimHumanoid"})
hAnimJoint = HAnimJoint = type("HAnimJoint", (Element,), {"name": "HAnimJoint"})
hAnimSegment = HAnimSegment = type("HAnimSegment", (Element,), {"name": "HAnimSegment"})
hAnimSite = HAnimSite = type("HAnimSite", (Element,), {"name": "HAnimSite"})
imageTexture = ImageTexture = type("ImageTexture", (Element,), {"name": "ImageTexture"})
imageTexture3D = ImageTexture3D = type("ImageTexture3D", (Element,), {"name": "ImageTexture3D"})
imageTextureAtlas = ImageTextureAtlas = type("ImageTextureAtlas", (Element,), {"name": "ImageTextureAtlas"})
indexedFaceSet = IndexedFaceSet = type("IndexedFaceSet", (Element,), {"name": "IndexedFaceSet"})
indexedLineSet = IndexedLineSet = type("IndexedLineSet", (Element,), {"name": "IndexedLineSet"})
indexedQuadSet = IndexedQuadSet = type("IndexedQuadSet", (Element,), {"name": "IndexedQuadSet"})
indexedTriangleSet = IndexedTriangleSet = type("IndexedTriangleSet", (Element,), {"name": "IndexedTriangleSet"})
indexedTriangleStripSet = IndexedTriangleStripSet = type(
    "IndexedTriangleStripSet", (Element,), {"name": "IndexedTriangleStripSet"}
)
isoSurfaceVolumeData = IsoSurfaceVolumeData = type("IsoSurfaceVolumeData", (Element,), {"name": "IsoSurfaceVolumeData"})
lineProperties = LineProperties = type("LineProperties", (Element,), {"name": "LineProperties"})
lineSet = LineSet = type("LineSet", (Element,), {"name": "LineSet"})
lOD = LOD = type("LOD", (Element,), {"name": "LOD"})
matrixTextureTransform = MatrixTextureTransform = type(
    "MatrixTextureTransform", (Element,), {"name": "MatrixTextureTransform"}
)
matrixTransform = MatrixTransform = type("MatrixTransform", (Element,), {"name": "MatrixTransform"})
mesh = Mesh = type("Mesh", (Element,), {"name": "Mesh"})
metadataBoolean = MetadataBoolean = type("MetadataBoolean", (Element,), {"name": "MetadataBoolean"})
metadataDouble = MetadataDouble = type("MetadataDouble", (Element,), {"name": "MetadataDouble"})
metadataFloat = MetadataFloat = type("MetadataFloat", (Element,), {"name": "MetadataFloat"})
metadataInteger = MetadataInteger = type("MetadataInteger", (Element,), {"name": "MetadataInteger"})
metadataSet = MetadataSet = type("MetadataSet", (Element,), {"name": "MetadataSet"})
metadataString = MetadataString = type("MetadataString", (Element,), {"name": "MetadataString"})
motorJoint = MotorJoint = type("MotorJoint", (Element,), {"name": "MotorJoint"})
movieTexture = MovieTexture = type("MovieTexture", (Element,), {"name": "MovieTexture"})
mPRPlane = MPRPlane = type("MPRPlane", (Element,), {"name": "MPRPlane"})
mPRVolumeStyle = MPRVolumeStyle = type("MPRVolumeStyle", (Element,), {"name": "MPRVolumeStyle"})
multiTexture = MultiTexture = type("MultiTexture", (Element,), {"name": "MultiTexture"})
multiTextureCoordinate = MultiTextureCoordinate = type(
    "MultiTextureCoordinate", (Element,), {"name": "MultiTextureCoordinate"}
)
navigationInfo = NavigationInfo = type("NavigationInfo", (Element,), {"name": "NavigationInfo"})
normal = Normal = type("Normal", (Element,), {"name": "Normal"})
normalInterpolator = NormalInterpolator = type("NormalInterpolator", (Element,), {"name": "NormalInterpolator"})
nozzle = Nozzle = type("Nozzle", (Element,), {"name": "Nozzle"})
opacityMapVolumeStyle = OpacityMapVolumeStyle = type(
    "OpacityMapVolumeStyle", (Element,), {"name": "OpacityMapVolumeStyle"}
)
orientationChaser = OrientationChaser = type("OrientationChaser", (Element,), {"name": "OrientationChaser"})
orientationDamper = OrientationDamper = type("OrientationDamper", (Element,), {"name": "OrientationDamper"})
orientationInterpolator = OrientationInterpolator = type(
    "OrientationInterpolator", (Element,), {"name": "OrientationInterpolator"}
)
orthoViewpoint = OrthoViewpoint = type("OrthoViewpoint", (Element,), {"name": "OrthoViewpoint"})
param = Param = type("Param", (Element,), {"name": "Param"})
particleSet = ParticleSet = type("ParticleSet", (Element,), {"name": "ParticleSet"})
physicalEnvironmentLight = PhysicalEnvironmentLight = type(
    "PhysicalEnvironmentLight", (Element,), {"name": "PhysicalEnvironmentLight"}
)
physicalMaterial = PhysicalMaterial = type("PhysicalMaterial", (Element,), {"name": "PhysicalMaterial"})
pixelTexture = PixelTexture = type("PixelTexture", (Element,), {"name": "PixelTexture"})
pixelTexture3D = PixelTexture3D = type("PixelTexture3D", (Element,), {"name": "PixelTexture3D"})
planeSensor = PlaneSensor = type("PlaneSensor", (Element,), {"name": "PlaneSensor"})
pointLight = PointLight = type("PointLight", (Element,), {"name": "PointLight"})
pointSet = PointSet = type("PointSet", (Element,), {"name": "PointSet"})
polyline2D = Polyline2D = type("Polyline2D", (Element,), {"name": "Polyline2D"})
polypoint2D = Polypoint2D = type("Polypoint2D", (Element,), {"name": "Polypoint2D"})
popGeometry = PopGeometry = type("PopGeometry", (Element,), {"name": "PopGeometry"})
popGeometryLevel = PopGeometryLevel = type("PopGeometryLevel", (Element,), {"name": "PopGeometryLevel"})
positionChaser = PositionChaser = type("PositionChaser", (Element,), {"name": "PositionChaser"})
positionChaser2D = PositionChaser2D = type("PositionChaser2D", (Element,), {"name": "PositionChaser2D"})
positionDamper = PositionDamper = type("PositionDamper", (Element,), {"name": "PositionDamper"})
positionDamper2D = PositionDamper2D = type("PositionDamper2D", (Element,), {"name": "PositionDamper2D"})
positionInterpolator = PositionInterpolator = type("PositionInterpolator", (Element,), {"name": "PositionInterpolator"})
positionInterpolator2D = PositionInterpolator2D = type(
    "PositionInterpolator2D", (Element,), {"name": "PositionInterpolator2D"}
)
projectionVolumeStyle = ProjectionVolumeStyle = type(
    "ProjectionVolumeStyle", (Element,), {"name": "ProjectionVolumeStyle"}
)
pyramid = Pyramid = type("Pyramid", (Element,), {"name": "Pyramid"})
quadSet = QuadSet = type("QuadSet", (Element,), {"name": "QuadSet"})
radarVolumeStyle = RadarVolumeStyle = type("RadarVolumeStyle", (Element,), {"name": "RadarVolumeStyle"})
rectangle2D = Rectangle2D = type("Rectangle2D", (Element,), {"name": "Rectangle2D"})
rectangularTorus = RectangularTorus = type("RectangularTorus", (Element,), {"name": "RectangularTorus"})
refinementTexture = RefinementTexture = type("RefinementTexture", (Element,), {"name": "RefinementTexture"})
remoteSelectionGroup = RemoteSelectionGroup = type("RemoteSelectionGroup", (Element,), {"name": "RemoteSelectionGroup"})
renderedTexture = RenderedTexture = type("RenderedTexture", (Element,), {"name": "RenderedTexture"})
rigidBody = RigidBody = type("RigidBody", (Element,), {"name": "RigidBody"})
rigidBodyCollection = RigidBodyCollection = type("RigidBodyCollection", (Element,), {"name": "RigidBodyCollection"})
scalarChaser = ScalarChaser = type("ScalarChaser", (Element,), {"name": "ScalarChaser"})
scalarDamper = ScalarDamper = type("ScalarDamper", (Element,), {"name": "ScalarDamper"})
scalarInterpolator = ScalarInterpolator = type("ScalarInterpolator", (Element,), {"name": "ScalarInterpolator"})
segmentedVolumeData = SegmentedVolumeData = type("SegmentedVolumeData", (Element,), {"name": "SegmentedVolumeData"})
shadedVolumeStyle = ShadedVolumeStyle = type("ShadedVolumeStyle", (Element,), {"name": "ShadedVolumeStyle"})
shaderPart = ShaderPart = type("ShaderPart", (Element,), {"name": "ShaderPart"})
silhouetteEnhancementVolumeStyle = SilhouetteEnhancementVolumeStyle = type(
    "SilhouetteEnhancementVolumeStyle",
    (Element,),
    {"name": "SilhouetteEnhancementVolumeStyle"},
)
singleAxisHingeJoint = SingleAxisHingeJoint = type("SingleAxisHingeJoint", (Element,), {"name": "SingleAxisHingeJoint"})
sliderJoint = SliderJoint = type("SliderJoint", (Element,), {"name": "SliderJoint"})
slopedCylinder = SlopedCylinder = type("SlopedCylinder", (Element,), {"name": "SlopedCylinder"})
snout = Snout = type("Snout", (Element,), {"name": "Snout"})
solidOfRevolution = SolidOfRevolution = type("SolidOfRevolution", (Element,), {"name": "SolidOfRevolution"})
sound = Sound = type("Sound", (Element,), {"name": "Sound"})
sphereSegment = SphereSegment = type("SphereSegment", (Element,), {"name": "SphereSegment"})
sphereSensor = SphereSensor = type("SphereSensor", (Element,), {"name": "SphereSensor"})
splinePositionInterpolator = SplinePositionInterpolator = type(
    "SplinePositionInterpolator", (Element,), {"name": "SplinePositionInterpolator"}
)
spotLight = SpotLight = type("SpotLight", (Element,), {"name": "SpotLight"})
staticGroup = StaticGroup = type("StaticGroup", (Element,), {"name": "StaticGroup"})
stippleVolumeStyle = StippleVolumeStyle = type("StippleVolumeStyle", (Element,), {"name": "StippleVolumeStyle"})
surfaceShaderTexture = SurfaceShaderTexture = type("SurfaceShaderTexture", (Element,), {"name": "SurfaceShaderTexture"})
switch = Switch = type("Switch", (Element,), {"name": "Switch"})
texCoordDamper2D = TexCoordDamper2D = type("TexCoordDamper2D", (Element,), {"name": "TexCoordDamper2D"})
text = Text = type("Text", (Element,), {"name": "Text"})
texture = Texture = type("Texture", (Element,), {"name": "Texture"})
textureCoordinate = TextureCoordinate = type("TextureCoordinate", (Element,), {"name": "TextureCoordinate"})
textureCoordinate3D = TextureCoordinate3D = type("TextureCoordinate3D", (Element,), {"name": "TextureCoordinate3D"})
textureCoordinateGenerator = TextureCoordinateGenerator = type(
    "TextureCoordinateGenerator", (Element,), {"name": "TextureCoordinateGenerator"}
)
textureProperties = TextureProperties = type("TextureProperties", (Element,), {"name": "TextureProperties"})
textureTransform = TextureTransform = type("TextureTransform", (Element,), {"name": "TextureTransform"})
textureTransform3D = TextureTransform3D = type("TextureTransform3D", (Element,), {"name": "TextureTransform3D"})
textureTransformMatrix3D = TextureTransformMatrix3D = type(
    "TextureTransformMatrix3D", (Element,), {"name": "TextureTransformMatrix3D"}
)
toneMappedVolumeStyle = ToneMappedVolumeStyle = type(
    "ToneMappedVolumeStyle", (Element,), {"name": "ToneMappedVolumeStyle"}
)
torus = Torus = type("Torus", (Element,), {"name": "Torus"})
touchSensor = TouchSensor = type("TouchSensor", (Element,), {"name": "TouchSensor"})
triangleSet = TriangleSet = type("TriangleSet", (Element,), {"name": "TriangleSet"})
triangleSet2D = TriangleSet2D = type("TriangleSet2D", (Element,), {"name": "TriangleSet2D"})
twoSidedMaterial = TwoSidedMaterial = type("TwoSidedMaterial", (Element,), {"name": "TwoSidedMaterial"})
uniform = Uniform = type("Uniform", (Element,), {"name": "Uniform"})
universalJoint = UniversalJoint = type("UniversalJoint", (Element,), {"name": "UniversalJoint"})
viewfrustum = Viewfrustum = type("Viewfrustum", (Element,), {"name": "Viewfrustum"})
viewpoint = Viewpoint = type("Viewpoint", (Element,), {"name": "Viewpoint"})
volumeData = VolumeData = type("VolumeData", (Element,), {"name": "VolumeData"})
worldInfo = WorldInfo = type("WorldInfo", (Element,), {"name": "WorldInfo"})
x3DAppearanceChildNode = X3DAppearanceChildNode = type(
    "X3DAppearanceChildNode", (Element,), {"name": "X3DAppearanceChildNode"}
)
x3DAppearanceNode = X3DAppearanceNode = type("X3DAppearanceNode", (Element,), {"name": "X3DAppearanceNode"})
x3DBackgroundNode = X3DBackgroundNode = type("X3DBackgroundNode", (Element,), {"name": "X3DBackgroundNode"})
x3DBinaryContainerGeometryNode = X3DBinaryContainerGeometryNode = type(
    "X3DBinaryContainerGeometryNode",
    (Element,),
    {"name": "X3DBinaryContainerGeometryNode"},
)
x3DBindableNode = X3DBindableNode = type("X3DBindableNode", (Element,), {"name": "X3DBindableNode"})
x3DBoundedObject = X3DBoundedObject = type("X3DBoundedObject", (Element,), {"name": "X3DBoundedObject"})
x3DChaserNode = X3DChaserNode = type("X3DChaserNode", (Element,), {"name": "X3DChaserNode"})
x3DChildNode = X3DChildNode = type("X3DChildNode", (Element,), {"name": "X3DChildNode"})
x3DColorNode = X3DColorNode = type("X3DColorNode", (Element,), {"name": "X3DColorNode"})
x3DComposableVolumeRenderStyleNode = X3DComposableVolumeRenderStyleNode = type(
    "X3DComposableVolumeRenderStyleNode",
    (Element,),
    {"name": "X3DComposableVolumeRenderStyleNode"},
)
x3DComposedGeometryNode = X3DComposedGeometryNode = type(
    "X3DComposedGeometryNode", (Element,), {"name": "X3DComposedGeometryNode"}
)
x3DCoordinateNode = X3DCoordinateNode = type("X3DCoordinateNode", (Element,), {"name": "X3DCoordinateNode"})
x3DDamperNode = X3DDamperNode = type("X3DDamperNode", (Element,), {"name": "X3DDamperNode"})
x3DDragSensorNode = X3DDragSensorNode = type("X3DDragSensorNode", (Element,), {"name": "X3DDragSensorNode"})
x3DEnvironmentNode = X3DEnvironmentNode = type("X3DEnvironmentNode", (Element,), {"name": "X3DEnvironmentNode"})
x3DEnvironmentTextureNode = X3DEnvironmentTextureNode = type(
    "X3DEnvironmentTextureNode", (Element,), {"name": "X3DEnvironmentTextureNode"}
)
x3DFogNode = X3DFogNode = type("X3DFogNode", (Element,), {"name": "X3DFogNode"})
x3DFollowerNode = X3DFollowerNode = type("X3DFollowerNode", (Element,), {"name": "X3DFollowerNode"})
x3DFontStyleNode = X3DFontStyleNode = type("X3DFontStyleNode", (Element,), {"name": "X3DFontStyleNode"})
x3DGeometricPropertyNode = X3DGeometricPropertyNode = type(
    "X3DGeometricPropertyNode", (Element,), {"name": "X3DGeometricPropertyNode"}
)
x3DGeometryNode = X3DGeometryNode = type("X3DGeometryNode", (Element,), {"name": "X3DGeometryNode"})
x3DGroupingNode = X3DGroupingNode = type("X3DGroupingNode", (Element,), {"name": "X3DGroupingNode"})
x3DInfoNode = X3DInfoNode = type("X3DInfoNode", (Element,), {"name": "X3DInfoNode"})
x3DInterpolatorNode = X3DInterpolatorNode = type("X3DInterpolatorNode", (Element,), {"name": "X3DInterpolatorNode"})
x3DLightNode = X3DLightNode = type("X3DLightNode", (Element,), {"name": "X3DLightNode"})
x3DLODNode = X3DLODNode = type("X3DLODNode", (Element,), {"name": "X3DLODNode"})
x3DMaterialNode = X3DMaterialNode = type("X3DMaterialNode", (Element,), {"name": "X3DMaterialNode"})
x3DMetadataObject = X3DMetadataObject = type("X3DMetadataObject", (Element,), {"name": "X3DMetadataObject"})
x3DNavigationInfoNode = X3DNavigationInfoNode = type(
    "X3DNavigationInfoNode", (Element,), {"name": "X3DNavigationInfoNode"}
)
x3DNBodyCollidableNode = X3DNBodyCollidableNode = type(
    "X3DNBodyCollidableNode", (Element,), {"name": "X3DNBodyCollidableNode"}
)
x3DNode = X3DNode = type("X3DNode", (Element,), {"name": "X3DNode"})
x3DPlanarGeometryNode = X3DPlanarGeometryNode = type(
    "X3DPlanarGeometryNode", (Element,), {"name": "X3DPlanarGeometryNode"}
)
x3DPointingDeviceSensorNode = X3DPointingDeviceSensorNode = type(
    "X3DPointingDeviceSensorNode", (Element,), {"name": "X3DPointingDeviceSensorNode"}
)
x3DRigidJointNode = X3DRigidJointNode = type("X3DRigidJointNode", (Element,), {"name": "X3DRigidJointNode"})
x3DSensorNode = X3DSensorNode = type("X3DSensorNode", (Element,), {"name": "X3DSensorNode"})
x3DShaderNode = X3DShaderNode = type("X3DShaderNode", (Element,), {"name": "X3DShaderNode"})
x3DShapeNode = X3DShapeNode = type("X3DShapeNode", (Element,), {"name": "X3DShapeNode"})
x3DSoundNode = X3DSoundNode = type("X3DSoundNode", (Element,), {"name": "X3DSoundNode"})
x3DSoundSourceNode = X3DSoundSourceNode = type("X3DSoundSourceNode", (Element,), {"name": "X3DSoundSourceNode"})
x3DSpatialGeometryNode = X3DSpatialGeometryNode = type(
    "X3DSpatialGeometryNode", (Element,), {"name": "X3DSpatialGeometryNode"}
)
x3DTexture3DNode = X3DTexture3DNode = type("X3DTexture3DNode", (Element,), {"name": "X3DTexture3DNode"})
x3DTextureCoordinateNode = X3DTextureCoordinateNode = type(
    "X3DTextureCoordinateNode", (Element,), {"name": "X3DTextureCoordinateNode"}
)
x3DTextureNode = X3DTextureNode = type("X3DTextureNode", (Element,), {"name": "X3DTextureNode"})
x3DTextureTransformNode = X3DTextureTransformNode = type(
    "X3DTextureTransformNode", (Element,), {"name": "X3DTextureTransformNode"}
)
x3DTimeDependentNode = X3DTimeDependentNode = type("X3DTimeDependentNode", (Element,), {"name": "X3DTimeDependentNode"})
x3DTouchSensorNode = X3DTouchSensorNode = type("X3DTouchSensorNode", (Element,), {"name": "X3DTouchSensorNode"})
x3DTransformNode = X3DTransformNode = type("X3DTransformNode", (Element,), {"name": "X3DTransformNode"})
x3DVertexAttributeNode = X3DVertexAttributeNode = type(
    "X3DVertexAttributeNode", (Element,), {"name": "X3DVertexAttributeNode"}
)
x3DViewpointNode = X3DViewpointNode = type("X3DViewpointNode", (Element,), {"name": "X3DViewpointNode"})
x3DVolumeDataNode = X3DVolumeDataNode = type("X3DVolumeDataNode", (Element,), {"name": "X3DVolumeDataNode"})
x3DVolumeRenderStyleNode = X3DVolumeRenderStyleNode = type(
    "X3DVolumeRenderStyleNode", (Element,), {"name": "X3DVolumeRenderStyleNode"}
)



"""
    domonic.d3.queue
    ====================================

    Small Python port of d3-queue's callback-oriented API.
"""

from __future__ import annotations

from typing import Any, Callable

from domonic.javascript import Error

Task = Callable[..., Any]


class Queue:
    def __init__(self, size: int | float):
        if size < 1:
            raise Error("invalid concurrency")

        self._size = size
        self._error: Exception | None = None
        self._tasks: list[tuple[Task, tuple[Any, ...]]] = []
        self._results: list[Any] = []
        self._call: Callable[[Exception | None, list[Any]], Any] | None = None
        self._started = False

    def defer(self, callback: Task, *args: Any) -> Queue:
        if not callable(callback):
            raise Error("invalid callback")
        if self._call is not None:
            raise Error("defer after await")
        if self._error is not None:
            return self

        self._tasks.append((callback, args))
        return self

    def abort(self) -> Queue:
        if self._error is None:
            self._error = Error("abort")
        self._notify()
        return self

    def await_(self, callback: Callable[..., Any]) -> Queue:
        if not callable(callback):
            raise Error("invalid callback")
        if self._call is not None:
            raise Error("multiple await")

        def _call(error: Exception | None, results: list[Any]) -> Any:
            return callback(error, *results)

        self._call = _call
        self._run()
        return self

    def awaitAll(self, callback: Callable[[Exception | None, list[Any]], Any]) -> Queue:
        if not callable(callback):
            raise Error("invalid callback")
        if self._call is not None:
            raise Error("multiple await")

        self._call = callback
        self._run()
        return self

    def _run(self) -> None:
        if self._started:
            self._notify()
            return

        self._started = True
        self._results = [None] * len(self._tasks)

        for index, (task, args) in enumerate(self._tasks):
            if self._error is not None:
                break

            finished = False

            def done(error: Exception | None = None, result: Any = None, *, _index: int = index) -> None:
                nonlocal finished
                if finished:
                    return
                finished = True
                if error is not None and self._error is None:
                    self._error = error
                    return
                self._results[_index] = result

            try:
                result = task(*args, done)
            except Exception as error:
                done(error)
                continue

            if not finished:
                done(None, result)

        self._notify()

    def _notify(self) -> None:
        if self._call is not None and self._started:
            callback = self._call
            self._call = None
            callback(self._error, self._results)


def queue(concurrency: int | float | None = None) -> Queue:
    if concurrency is None:
        concurrency = float("inf")
    return Queue(concurrency)



import unittest

from domonic.constants import (
    HTTPStatus,
    doctypes,
    file_extensions,
    get_doctype,
    get_mime_type,
    get_namespace,
    get_status_text,
    http_response_status_codes,
    mime_types,
    namespaces,
)
from domonic.constants.color import Color
from domonic.constants.entities import Char, Entity
from domonic.constants.keyboard import Code, Key, KeyCode, KeyLocation, normalize_code, normalize_key


class ConstantsTest(unittest.TestCase):
    def test_namespace_constants(self):
        self.assertEqual(namespaces["svg"], "http://www.w3.org/2000/svg")
        self.assertEqual(namespaces["html"], "http://www.w3.org/1999/xhtml")
        self.assertEqual(namespaces["atom"], "http://www.w3.org/2005/Atom")
        self.assertEqual(namespaces["news"], "http://www.google.com/schemas/sitemap-news/0.9")
        self.assertEqual(namespaces["video"], "http://www.google.com/schemas/sitemap-video/1.1")
        self.assertEqual(get_namespace("svg"), "http://www.w3.org/2000/svg")
        self.assertEqual(get_namespace("missing", "fallback"), "fallback")

    def test_doctype_constants(self):
        self.assertEqual(doctypes["HTML5"], "<!DOCTYPE html>")
        self.assertIn("HTML 4.01", doctypes["HTML4_01_Strict"])
        self.assertIn("XHTML 1.0 Strict", doctypes["XHTML1_0_Strict"])
        self.assertIn("svg", doctypes["SVG1_1"].lower())
        self.assertEqual(get_doctype("HTML5"), "<!DOCTYPE html>")
        self.assertEqual(get_doctype("missing", "fallback"), "fallback")

    def test_http_status_constants(self):
        self.assertEqual(int(HTTPStatus.OK), 200)
        self.assertEqual(http_response_status_codes[404], "Not Found")
        self.assertEqual(http_response_status_codes[418], "I'm a Teapot")
        self.assertEqual(http_response_status_codes[200], "OK")
        self.assertEqual(get_status_text(200), "OK")
        self.assertEqual(get_status_text(499), "Client Closed Request")
        self.assertEqual(get_status_text(999, "fallback"), "fallback")

    def test_file_extension_constants(self):
        self.assertEqual(file_extensions["svg"], "image/svg+xml")
        self.assertEqual(file_extensions["json"], "application/json")
        self.assertEqual(file_extensions["wasm"], "application/wasm")
        self.assertEqual(file_extensions["webp"], "image/webp")
        self.assertEqual(file_extensions["woff2"], "font/woff2")
        self.assertEqual(file_extensions["csv"], "text/csv")
        self.assertIs(mime_types, file_extensions)
        self.assertEqual(get_mime_type(".svg"), "image/svg+xml")
        self.assertEqual(get_mime_type("JSON"), "application/json")
        self.assertEqual(get_mime_type("wasm"), "application/wasm")
        self.assertEqual(get_mime_type(".woff2"), "font/woff2")
        self.assertEqual(get_mime_type("missing", "fallback"), "fallback")

    def test_keyboard_modifier_helper(self):
        self.assertTrue(KeyCode.is_modifier(KeyCode.SHIFT))
        self.assertTrue(KeyCode.is_modifier(KeyCode.LEFT_COMMAND))
        self.assertFalse(KeyCode.is_modifier(KeyCode.ENTER))

    def test_modern_keyboard_constants(self):
        self.assertEqual(Key.ENTER, "Enter")
        self.assertEqual(Key.SPACE, " ")
        self.assertEqual(Code.KEY_A, "KeyA")
        self.assertEqual(Code.DIGIT_7, "Digit7")
        self.assertEqual(KeyLocation.STANDARD, 0)
        self.assertEqual(KeyLocation.LEFT, 1)
        self.assertEqual(KeyLocation.RIGHT, 2)
        self.assertEqual(KeyLocation.NUMPAD, 3)
        self.assertEqual(normalize_key("Esc"), "Escape")
        self.assertEqual(normalize_key("A"), "a")
        self.assertEqual(normalize_code("a"), "KeyA")
        self.assertEqual(normalize_code("7"), "Digit7")
        self.assertEqual(KeyCode.to_key("13"), "Enter")
        self.assertEqual(KeyCode.to_code("65"), "KeyA")
        self.assertEqual(KeyCode.from_key("Enter"), 13)

    def test_entity_and_char_rendering(self):
        self.assertEqual(str(Entity("&amp;")), "&")
        self.assertEqual(repr(Entity("&lt;")), "Entity('&lt;')")
        self.assertEqual(str(Char("<tag>")), "&lt;tag&gt;")

    def test_entity_disambiguated_aliases(self):
        self.assertEqual(Char.AGRAVE, "&agrave;")
        self.assertEqual(Char.LATIN_CAPITAL_A_GRAVE, "&Agrave;")
        self.assertEqual(Char.ALPHA, "&alpha;")
        self.assertEqual(Char.GREEK_CAPITAL_ALPHA, "&Alpha;")
        self.assertEqual(Char.DAGGER, "&Dagger;")
        self.assertEqual(Char.SINGLE_DAGGER, "&dagger;")
        self.assertEqual(Char.DOUBLE_DAGGER, "&Dagger;")
        self.assertEqual(Char.PRIME, "&Prime;")
        self.assertEqual(Char.SINGLE_PRIME, "&prime;")
        self.assertEqual(Char.DOUBLE_PRIME, "&Prime;")
        self.assertEqual(Char.TILDE, "&tilde;")
        self.assertEqual(Char.ASCII_TILDE, "&#126;")
        self.assertEqual(Char.YUML, "&Yuml;")
        self.assertEqual(Char.LATIN_SMALL_Y_UMLAUT, "&yuml;")

    def test_color_helpers(self):
        self.assertEqual(Color.hex2rgb("#ffffff"), (255, 255, 255))
        self.assertEqual(Color.rgb2hex(255, 0, 255), "#ff00ff")
        self.assertEqual(Color(255, 0, 255).toRGBA(), (255, 0, 255, 1))
        self.assertEqual(Color("#00ff00").convert("css"), "#00ff00")


if __name__ == "__main__":
    unittest.main()



"""
    test_cmd
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.cmd import dir
from domonic.utils import Utils

# from domonic.decorators import silence


class TestCase(unittest.TestCase):
    def test_cmd_dir(self):
        if not Utils.is_windows():
            return

        files = dir()
        print(files)
        # return
        self.assertIn("domonic", str(dir()))

        try:
            print(dir("..\\"))  # notice requires 2 backslashes
        except Exception:
            # will fail on non windows
            pass

        for line in dir():
            print("line:", line)


if __name__ == "__main__":
    unittest.main()



"""
    domonic.sitemap
    ====================================

    generate or load sitemaps

    warning - when using image and video tags from this package they will be namespaced i.e <image:image> and <video:video>
    so i'd advise to only import them within the def that you use them in to avoid conflict with html.image

"""

import datetime

from domonic.dom import Document, Element

# __all__ = ['sitemap', 'url', 'lastmod']

sitemap_tags = [
    "sitemapindex",
    "sitemap",
    "urlset",
    "url",
    "loc",
    "lastmod",
    "changefreq",
    "priority",
    "image:image",
    "image:loc",
    "image:caption",
    "image:title",
    "image:geo_location",
    "image:license",
    "video:video",
    "video:loc",
    "video:caption",
    "video:title",
    "video:thumbnail_loc",
    "geo:geo",
    "geo:location",
    "atom:entry",
    "atom:link",
    "xhtml:link",
    "mobile:mobile"
]

sitemap_attributes = [
    "xmlns",
    "xmlns:xsi",
    "xsi:schemaLocation",
    "xmlns:xhtml",
    "xmlns:xlink",
    "xmlns:atom",
    "xmlns:geo",
]
# sitemap_change_frequencies = ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']

XMLNS: str = "http://www.sitemaps.org/schemas/sitemap/0.9"
XMLNS_XSI: str = "http://www.w3.org/2001/XMLSchema-instance"
SCHEMA_SITEINDEX: str = "http://www.sitemaps.org/schemas/sitemap/0.9/siteindex.xsd"
SCHEMA_SITEMAP: str = "http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd"
#     xmlns="http://www.google.com/schemas/sitemap-image/1.1"
#     xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">


def sitemap_format(self, *args, **kwargs):
    """attempts to prettify the output of the sitemap."""
    outp = f"<{self.name}{self.__attributes__}>{self.content}</{self.name}>"
    from domonic.xml import prettify

    return prettify(outp)


sitemapindex = type(
    "sitemapindex",
    (Document,),
    {
        "name": "sitemapindex",
        "xmlns": XMLNS,
        "xmlns:xsi": XMLNS_XSI,
        "xsi:schemaLocation": SCHEMA_SITEINDEX,
        "__format__": sitemap_format,
    },
)

sitemap = type("sitemap", (Element,), {"name": "sitemap"})

urlset = type(
    "urlset",
    (Element,),
    {
        "name": "urlset",
        "xmlns:xsi": XMLNS_XSI,
        "xsi:schemaLocation": SCHEMA_SITEMAP,
        "xmlns": XMLNS,
    },
)


url = type("url", (Element,), {"name": "url"})
loc = type("loc", (Element,), {"name": "loc"})
lastmod = type("lastmod", (Element,), {"name": "lastmod"})
changefreq = type("changefreq", (Element,), {"name": "changefreq"})
priority = type("priority", (Element,), {"name": "priority"})


def sitemapindex_from_urls(urls):
    """
    Create a sitemap index from a list of urls.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    # i.e

    # <?xml version="1.0" encoding="UTF-8"?>
    # <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
    # 	<sitemap>
    # 		<loc>https://xyz.com/sitemap1.xml</loc>
    # 		<lastmod>2021-07-08T13:12:16+00:00</lastmod>
    # 	</sitemap>
    # </sitemapindex>

    """
    sitemap_index = sitemapindex()
    for url in urls:
        sitemap_index.append(sitemap(loc(url), lastmod(datetime.datetime.now())))


def sitemap_from_urls(urls):
    """
    Create a sitemap from a list of urls.add()

    Note: This won't allow you to add priority or changefreq of the urls. or add images etc
    tho u could loop the nodes afterwards and do that.

    WARNING:
    there's a difference between a sitemap index and a sitemap.
    make sure you know what you want.

    """
    sitemap = urlset()
    for url in urls:
        sitemap.append(url(loc(url), lastmod(datetime.datetime.now())))
    return sitemap


def get_sitemap(path: str, *args, **kwargs):
    """
    Download a sitemap
    """
    import domonic

    sitemap_text = _get_sitemap_text(path)

    try:
        return domonic.domonic.parseString(sitemap_text, parser="expat")
    except Exception:
        pass

    try:
        return domonic.domonic.parseString(sitemap_text, parser="html5_parser")
    except Exception:
        pass

    return domonic.domonic.parseString(sitemap_text)


def _get_sitemap_text(path: str) -> str:
    import requests

    r = requests.get(path)
    return r.text


# image
image_image = globals()["image:image"] = type("image:image", (Element,), {"name": "image:image", "ns": "image"})
image_loc = globals()["image:loc"] = type("image:loc", (Element,), {"name": "image:loc", "ns": "image"})
image_caption = globals()["image:caption"] = type("image:caption", (Element,), {"name": "image:caption", "ns": "image"})
image_title = globals()["image:title"] = type("image:title", (Element,), {"name": "image:title", "ns": "image"})
image_geo_location = globals()["image:geo_location"] = type("image:geo_location", (Element,), {"name": "image:geo_location", "ns": "image"})
image_license = globals()["image:license"] = type("image:license", (Element,), {"name": "image:license", "ns": "image"})

# video
video_video = globals()["video:video"] = type("video:video", (Element,), {"name": "video:video", "ns": "video"})
video_content_loc = globals()["video:content_loc"] = type("video:content_loc", (Element,), {"name": "video:content_loc", "ns": "video"})
video_thumbnail_loc = globals()["video:thumbnail_loc"] = type("video:thumbnail_loc", (Element,), {"name": "video:thumbnail_loc", "ns": "video"})
video_title = globals()["video:title"] = type("video:title", (Element,), {"name": "video:title", "ns": "video"})
video_description = globals()["video:description"] = type("video:description", (Element,), {"name": "video:description", "ns": "video"})
video_duration = globals()["video:duration"] = type("video:duration", (Element,), {"name": "video:duration", "ns": "video"})
video_publication_date = globals()["video:publication_date"] = type("video:publication_date", (Element,), {"name": "video:publication_date", "ns": "video"})
video_tags = globals()["video:tags"] = type("video:tags", (Element,), {"name": "video:tags", "ns": "video"})
video_category = globals()["video:category"] = type("video:category", (Element,), {"name": "video:category", "ns": "video"})
video_rating = globals()["video:rating"] = type("video:rating", (Element,), {"name": "video:rating", "ns": "video"})
video_view_count = globals()["video:view_count"] = type("video:view_count", (Element,), {"name": "video:view_count", "ns": "video"})
video_price = globals()["video:price"] = type("video:price", (Element,), {"name": "video:price", "ns": "video"})
video_price_currency = globals()["video:price_currency"] = type("video:price_currency", (Element,), {"name": "video:price_currency", "ns": "video"})

# news
news_news = globals()["news:news"] = type("news:news", (Element,), {"name": "news:news", "ns": "news"})
news_publication_date = globals()["news:publication_date"] = type("news:publication_date", (Element,), {"name": "news:publication_date", "ns": "news"})
news_title = globals()["news:title"] = type("news:title", (Element,), {"name": "news:title", "ns": "news"})
news_keywords = globals()["news:keywords"] = type("news:keywords", (Element,), {"name": "news:keywords", "ns": "news"})
news_stock_tickers = globals()["news:stock_tickers"] = type("news:stock_tickers", (Element,), {"name": "news:stock_tickers", "ns": "news"})

# geo
geo_geo = globals()["geo:geo"] = type("geo:geo", (Element,), {"name": "geo:geo", "ns": "geo"})
geo_place_name = globals()["geo:place_name"] = type("geo:place_name", (Element,), {"name": "geo:place_name", "ns": "geo"})
geo_country = globals()["geo:country"] = type("geo:country", (Element,), {"name": "geo:country", "ns": "geo"})

# atom
atom_link = globals()["atom:link"] = type("atom:link", (Element,), {"name": "atom:link", "ns": "atom"})

# xhtml
xhtml_link = globals()["xhtml:link"] = type("xhtml:link", (Element,), {"name": "xhtml:link", "ns": "xhtml"})

# mobile
mobile_mobile = globals()["mobile:mobile"] = type("mobile:mobile", (Element,), {"name": "mobile:mobile", "ns": "mobile"})

def create_ns_element(tag_name, **attributes):
    """Factory function to create elements dynamically."""
    if tag_name in globals():
        return globals()[tag_name](**attributes)
    raise ValueError(f"Tag '{tag_name}' is not defined in globals().")



import unittest

from domonic.animation import (
    Animation,
    AnimationPlaybackEvent,
    ComputedEffectTiming,
    EffectTiming,
    KeyframeEffect,
)
from domonic.dom import DocumentTimeline
from domonic.html import div


class TestCase(unittest.TestCase):
    def test_effect_timing_and_computed_effect_timing(self):
        timing = EffectTiming(duration=200, delay=50, iterations=2, fill="forwards", easing="ease-in")
        effect = KeyframeEffect(div(), [{"opacity": 0, "offset": 0}, {"opacity": 1, "offset": 1}], timing)

        raw = effect.getTiming()
        computed = effect.getComputedTiming(150)

        self.assertEqual(raw.duration, 200)
        self.assertEqual(raw.delay, 50)
        self.assertEqual(raw.iterations, 2)
        self.assertIsInstance(computed, ComputedEffectTiming)
        self.assertEqual(computed.activeDuration, 400)
        self.assertEqual(computed.endTime, 450)
        self.assertIsNotNone(computed.progress)

    def test_keyframe_effect_applies_interpolated_style_values(self):
        target = div()
        effect = KeyframeEffect(
            target,
            [
                {"offset": 0, "opacity": 0, "width": "100px"},
                {"offset": 1, "opacity": 1, "width": "200px"},
            ],
            {"duration": 1000, "fill": "forwards"},
        )

        effect.apply(500)

        self.assertEqual(target.style.width, "150px")
        self.assertAlmostEqual(float(target.style.opacity), 0.5)

    def test_animation_and_element_animate_surface(self):
        target = div()
        animation = target.animate(
            [{"offset": 0, "opacity": 0}, {"offset": 1, "opacity": 1}],
            {"duration": 1000, "fill": "forwards"},
        )

        self.assertIsInstance(animation, Animation)
        self.assertIsInstance(animation.timeline, DocumentTimeline)
        self.assertEqual(animation.playState, "running")

        animation.currentTime = 250
        self.assertAlmostEqual(float(target.style.opacity), 0.25)

        animation.updatePlaybackRate(2)
        self.assertEqual(animation.playbackRate, 2.0)

    def test_animation_finish_cancel_and_playback_events(self):
        target = div()
        effect = KeyframeEffect(target, [{"offset": 0, "opacity": 0}, {"offset": 1, "opacity": 1}], {"duration": 100})
        animation = Animation(effect, DocumentTimeline())
        events = []

        animation.addEventListener("finish", lambda event: events.append(("finish", event.currentTime)))
        animation.addEventListener("cancel", lambda event: events.append(("cancel", event.currentTime)))

        animation.play()
        animation.finish()
        self.assertEqual(animation.playState, "finished")
        self.assertEqual(target.style.opacity, 1)
        self.assertIsInstance(AnimationPlaybackEvent("finish"), AnimationPlaybackEvent)

        animation.cancel()
        self.assertEqual(animation.playState, "idle")
        self.assertEqual(events[0][0], "finish")
        self.assertEqual(events[1], ("cancel", None))


if __name__ == "__main__":
    unittest.main()



"""
    domonic.ext.justhtml_
    ====================================

    Adapter for using justhtml as a frontend parser while rebuilding a
    domonic document tree from the parsed output.
"""

from __future__ import annotations

from html5lib import HTMLParser

from domonic.ext.html5lib_ import getTreeBuilder


def _serialize(node) -> str:
    for attr in ("to_html", "serialize", "html", "to_string"):
        candidate = getattr(node, attr, None)
        if callable(candidate):
            try:
                return candidate(pretty=False)
            except TypeError:
                return candidate()
        if isinstance(candidate, str):
            return candidate
    return str(node)


def parse(html, return_root=True, **kwargs):
    from justhtml import JustHTML

    parsed = JustHTML(html, fragment=False, sanitize=False)
    serialized = _serialize(parsed)
    parser = HTMLParser(tree=getTreeBuilder())
    document = parser.parse(serialized)
    return document.documentElement if return_root else document



"""
    test_domonic
    ~~~~~~~~~~~~
    unit tests for JSON helpers
"""

import json
import os
import tempfile
import unittest

import domonic.JSON as JSON  # do this to use same way as previous versions of domonic
from domonic.decorators import as_json
from domonic.html import Element


class TestCase(unittest.TestCase):
    SOMEJSON = """
    { "items": [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        }
    ]}
    """
    SOMEJSON2 = """
    [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        },
        {
            "id":"08",
            "name": "DDD",
            "thing":{},
            "list":[],
            "extra":23
        }
    ]
    """

    def test_tablify(self):
        table_from_string = JSON.tablify(TestCase.SOMEJSON2)
        self.assertIsInstance(table_from_string, Element)
        self.assertEqual(table_from_string.tagName, "table")

        table_from_dict = JSON.tablify({"id": 1, "name": "test"})
        self.assertIsInstance(table_from_dict, Element)
        self.assertEqual(table_from_dict.tagName, "table")

        table_from_list = JSON.tablify(JSON.parse(TestCase.SOMEJSON2))
        self.assertEqual(table_from_list.getElementsByTagName("td")[0].textContent, "01")

        nested_table = JSON.tablify(JSON.parse(TestCase.SOMEJSON)["items"])
        self.assertEqual(nested_table.getElementsByTagName("td")[0].textContent, "01")
        self.assertIn("extra", str(table_from_string))

        with self.assertRaises(ValueError):
            JSON.tablify([1, 2, 3])

    def test_table2json(self):
        table_node = JSON.tablify(
            [
                {"id": "01", "name": "Java"},
                {"id": "07", "name": "C++"},
            ]
        )
        self.assertEqual(
            JSON.table2json(table_node),
            [
                {"id": "01", "name": "Java"},
                {"id": "07", "name": "C++"},
            ],
        )

        with self.assertRaises(ValueError):
            JSON.table2json("not-a-table")

    def test_parse_and_parse_file(self):
        parsed = JSON.parse(TestCase.SOMEJSON)
        self.assertEqual(parsed["items"][1]["name"], "C++")

        parsed_bytes = JSON.parse(TestCase.SOMEJSON.encode("utf-8"))
        self.assertEqual(parsed_bytes["items"][0]["id"], "01")

        with tempfile.NamedTemporaryFile("w+", suffix=".json", delete=False) as handle:
            handle.write('{"hello": "world"}')
            temp_path = handle.name

        try:
            self.assertEqual(JSON.parse_file(temp_path), {"hello": "world"})
        finally:
            os.unlink(temp_path)

    def test_stringify(self):
        payload = {"hi": [1, 2, 3]}
        text = JSON.stringify(payload, sort_keys=True)
        self.assertEqual(text, '{"hi": [1, 2, 3]}')

        with tempfile.NamedTemporaryFile("r", suffix=".json", delete=False) as handle:
            temp_path = handle.name

        try:
            written = JSON.stringify(payload, filepath=temp_path, sort_keys=True)
            self.assertEqual(written, '{"hi": [1, 2, 3]}')
            with open(temp_path, encoding="utf-8") as saved_file:
                self.assertEqual(saved_file.read(), written)
        finally:
            os.unlink(temp_path)

    def test_csvify_and_csv2json(self):
        data = [
            {"id": "01", "name": "Java"},
            {"name": "C++", "id": "07", "extra": "yes"},
        ]

        with tempfile.NamedTemporaryFile("r", suffix=".csv", delete=False) as handle:
            csv_path = handle.name

        try:
            returned_path = JSON.csvify(data, csv_path)
            self.assertEqual(returned_path, csv_path)
            with open(csv_path, encoding="utf-8") as csv_file:
                csv_text = csv_file.read()
            self.assertIn("id,name,extra", csv_text)
            self.assertIn("01,Java,", csv_text)
            self.assertIn("07,C++,yes", csv_text)

            payload = JSON.csv2json(csv_path)
            self.assertEqual(json.loads(payload)[1]["extra"], "yes")
        finally:
            os.unlink(csv_path)

    def test_flatten(self):
        flattened = JSON.flatten({"user": {"name": "domonic", "meta": {"role": "admin"}}})
        self.assertEqual(
            flattened,
            {"user__name": "domonic", "user__meta__role": "admin"},
        )

    def test_is_json(self):
        self.assertTrue(JSON.is_json('{"a": 1}'))
        self.assertTrue(JSON.is_json("[1, 2, 3]"))
        self.assertTrue(JSON.is_json(" true "))
        self.assertFalse(JSON.is_json("not json"))
        self.assertFalse(JSON.is_json(""))
        self.assertFalse(JSON.is_json(123))

    def test_as_json_decorator(self):
        @as_json
        def yo():
            return {"hi": [1, 2, 3]}

        self.assertEqual(yo(), '{"hi": [1, 2, 3]}')


if __name__ == "__main__":
    unittest.main()



"""Packaging configuration for domonic."""

import os
import re

from setuptools import find_packages, setup

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


def read(filename: str) -> str:
    """Returns the contents of a file.

    Args:
        filename (str): The name of the file to read.

    Returns:
        str: content of the file
    """
    with open(os.path.join(BASE_DIR, filename), encoding="utf-8") as file:
        return file.read()


def get_version() -> str:
    """Read the package version without importing domonic during setup."""
    version_file = read(os.path.join("domonic", "__init__.py"))
    match = re.search(r'^__version__\s*=\s*"([^"]+)"', version_file, re.MULTILINE)
    if not match:
        raise RuntimeError("Unable to find package version.")
    return match.group(1)


def get_requirements(filename: str = "requirements.txt"):
    """returns a list of all requirements"""
    requirements = read(filename)
    return list(
        filter(
            None,
            [req.strip() for req in requirements.split() if not req.startswith("#")],
        )
    )

setup(
    name="domonic",
    version=get_version(),
    author="byteface",
    author_email="byteface@gmail.com",
    license="MIT",
    url="https://github.com/byteface/domonic",
    project_urls={
        "Documentation": "https://domonic.readthedocs.io/",
        "Source": "https://github.com/byteface/domonic",
        "Tracker": "https://github.com/byteface/domonic/issues",
        "Examples": "https://github.com/byteface/domonic/tree/master/examples",
        "Releases": "https://github.com/byteface/domonic/releases",
    },
    description="A Python DOM far beyond minidom, with HTML, SVG, events, web APIs, and a JavaScript-like runtime.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords=[
        "html",
        "generate",
        "templating",
        "dom",
        "vdom",
        "terminal",
        "json",
        "web",
        "template",
        "javascript",
        "DOM",
        "GUI",
        "render",
        "website",
        "apps",
        "html5",
        "framework",
        "SVG",
        "x3d",
        "events",
        "geom",
        "whatwg",
        "web api",
        "custom elements",
        "shadow dom",
        "css selectors",
        "html parser",
        "animation",
        "dom manipulation",
    ],
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: JavaScript",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: Terminals",
        "Topic :: Utilities",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
    ],
    install_requires=get_requirements(),
    packages=find_packages(exclude=("tests", "tests.*")),
    include_package_data=True,
    license_files=("LICENSE.txt",),
    entry_points={
        "console_scripts": [
            "domonic = domonic.__main__:run",
        ],
    },
)
